# flake8: noqa E501
__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import os
import shutil
import sys


def _check_if_should_setup() -> bool:
    """Check whether to connect to debug server or not"""
    return "pytest" not in sys.modules and not os.environ.get("QGIS_PLUGIN_IN_CI")


def setup_pydevd(host: str = "localhost", port: int = 5678) -> bool:
    """
    Setup pydevd debugging service

    Here is a sample (GlobeBuilder) Intellij Idea / PyCharm configuration for setting up the debug server in workspace.xml:

    <configuration name="Debug Server" type="PyRemoteDebugConfigurationType" factoryName="Python Remote Debug">
      <module name="QGIS Debug Server" />
      <option name="PORT" value="5678" />
      <option name="HOST" value="localhost" />
      <PathMappingSettings>
        <option name="pathMappings">
          <list>
            <mapping local-root="$PROJECT_DIR$/<plugin_name>" remote-root="/home/user/.local/share/QGIS/QGIS3/profiles/default/python/plugins/<plugin_name>" />
          </list>
        </option>
      </PathMappingSettings>
      <option name="REDIRECT_OUTPUT" value="true" />
      <option name="SUSPEND_AFTER_CONNECT" value="true" />
      <method v="2" />
    </configuration>

    :param host: host of the debug server
    :param port: port of the debug server
    :return: Whether debugger was initialized properly or not
    """
    succeeded = False
    if _check_if_should_setup():
        try:
            import pydevd

            pydevd.settrace(host, port=port, stdoutToServer=True, stderrToServer=True)
            succeeded = True
        except Exception as e:
            print("Unable to create pydevd debugger: {}".format(e))

    return succeeded


def setup_ptvsd(host: str = "localhost", port: int = 5678) -> bool:
    """
    Setup ptvsd degugging service

    Here is a sample VSCode configuration for connecting to the debug server in launch.json:

    {
    "version": "0.2.0",
    "configurations": [
        {
            "name": "Python: Remote Attach",
            "type": "python",
            "request": "attach",
            "connect": {
                "host": "localhost",
                "port": 5678
            },
            "pathMappings": [
                {
                    "localRoot": "${workspaceFolder}/<plugin_name>",
                    "remoteRoot": "/Users/<user_name>/Library/Application Support/QGIS/QGIS3/profiles/edplanning/python/plugins/<plugin_name>"
                }
            ]
        }
      ]
    }

    :param host: host of the debug server
    :param port: port of the debug server
    :return: Whether debugger was initialized properly or not
    """
    succeeded = False
    if _check_if_should_setup():
        try:
            import ptvsd

            ptvsd.enable_attach((host, port))
            succeeded = True
        except Exception as e:
            print("Unable to create ptvsd debugger: {}".format(e))
    return succeeded


def setup_debugpy(host: str = "localhost", port: int = 5678) -> bool:
    """
    Setup debugpy debugging service

    Here is a sample VSCode configuration for connecting to the debug server in launch.json:

    {
    "version": "0.2.0",
    "configurations": [
        {
            "name": "Python: Remote Attach",
            "type": "python",
            "request": "attach",
            "connect": {
                "host": "localhost",
                "port": 5678
            },
            "pathMappings": [
                {
                    "localRoot": "${workspaceFolder}/<plugin_name>",
                    "remoteRoot": "/Users/<user_name>/Library/Application Support/QGIS/QGIS3/profiles/edplanning/python/plugins/<plugin_name>"
                }
            ]
        }
      ]
    }

    :param host: host of the debug server
    :param port: port of the debug server
    :return: Whether debugger was initialized properly or not
    """
    succeeded = False
    if _check_if_should_setup() and not os.environ.get("QGIS_DEBUGPY_HAS_LOADED"):
        try:
            import debugpy

            debugpy.configure(python=shutil.which("python"))
            debugpy.listen((host, port))
            succeeded = True
        except Exception as e:
            print("Unable to create debugpy debugger: {}".format(e))
        else:
            # extra guard for debugpy not to setup it twice
            # (causes debugging session to hang)
            os.environ["QGIS_DEBUGPY_HAS_LOADED"] = "1"

    return succeeded



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import time

from qgis.core import Qgis

from ..testing.utilities import TestTaskRunner
from ..tools.custom_logging import bar_msg
from ..tools.decorations import log_if_fails, taskify
from ..tools.exceptions import QgsPluginNotImplementedException


@log_if_fails
def function_that_fails(arg, arg2, kwarg1=None, kwarg2=None):
    raise ValueError("Error message")


@log_if_fails
def function_that_shows_msg():
    raise QgsPluginNotImplementedException("Error message", bar_msg("Please implement"))


@taskify
def function_that_runs_as_a_task(arg, kwarg=None):
    for _ in range(10):
        time.sleep(0.01)
    return arg, kwarg


class MockClass:
    @log_if_fails
    def method_that_fails(self, arg, arg2, kwarg1=None, kwarg2=None):
        raise ValueError("M: Error message")

    @log_if_fails
    def method_that_shows_msg(self):
        raise QgsPluginNotImplementedException(
            "M: Error message", bar_msg("Please implement")
        )

    @taskify
    def method_that_runs_as_a_task(self):
        for _ in range(10):
            time.sleep(0.01)
        return True


def test_logging_if_fails(initialize_logger, qgis_iface):
    function_that_shows_msg()
    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)
    assert "Error message:Please implement" in messages


def test_logging_if_fails_method(initialize_logger, qgis_iface):
    MockClass().method_that_shows_msg()

    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)
    assert "M: Error message:Please implement" in messages


def test_logging_if_fails_without_details(initialize_logger, qgis_iface):
    function_that_fails(1, 2, 3, kwarg2=4)

    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)
    assert "Unhandled exception occurred:Error message" in messages


def test_logging_if_fails_without_details_method(initialize_logger, qgis_iface):
    MockClass().method_that_fails(1, 2, 3, kwarg2=4)

    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)
    assert "Unhandled exception occurred:M: Error message" in messages


def test_taskify(task_runner: TestTaskRunner):
    task = function_that_runs_as_a_task(1, kwarg=2)
    success = task_runner.run_task(task)

    assert success
    assert task.result == (1, 2)


def test_taskify_method(task_runner: TestTaskRunner):
    task = MockClass().method_that_runs_as_a_task()
    success = task_runner.run_task(task)

    assert success
    assert task.result is True



import time

from qgis.core import Qgis
from qgis.PyQt.QtCore import QCoreApplication, QEventLoop

from ..testing.utilities import SimpleTask, TestTaskRunner
from ..tools.exceptions import QgsPluginException, TaskInterruptedException
from ..tools.tasks import FunctionTask

__copyright__ = "Copyright 2021, qgis_plugin_tools contributors"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"


def fn(*args, **kwargs):  # noqa
    for _ in range(10):
        time.sleep(0.01)
    return args, kwargs


def test_run_simple_task(task_runner: TestTaskRunner):
    task = SimpleTask()
    success = task_runner.run_task(task)

    assert success
    assert task_runner.progress == 100


def test_run_simple_task_canceled(task_runner: TestTaskRunner, qgis_iface):
    task = SimpleTask()
    success = task_runner.run_task(task, cancel=True)

    # for some reason this randomly fails if the signal is not waited for
    QCoreApplication.processEvents(QEventLoop.AllEvents, 100)  # 100 ms wait

    messages = qgis_iface.messageBar().get_messages(Qgis.Warning)

    assert not success
    assert task_runner.fail
    assert (
        "Task SimpleTask was not successful:"
        "Task was cancelled by user or some dependency tasks failed" in messages
    )


def test_run_simple_task_canceled_after_a_while(
    task_runner: TestTaskRunner, qgis_iface
):
    task = SimpleTask()
    success = task_runner.run_task(task, cancel=True, sleep_before_cancel=0.01)
    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)

    assert not success
    assert task_runner.fail
    assert isinstance(task.exception, TaskInterruptedException)
    assert "Task canceled!:" in messages


def test_run_simple_task_failed(task_runner: TestTaskRunner, qgis_iface):
    task = SimpleTask(True)
    success = task_runner.run_task(task)
    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)

    assert not success
    assert task_runner.fail
    assert "Unhandled exception occurred:custom failure" in messages


def test_run_simple_task_failed_with_qgs_plugin_exception(
    task_runner: TestTaskRunner, qgis_iface
):
    task = SimpleTask(True, QgsPluginException)
    success = task_runner.run_task(task)
    messages = qgis_iface.messageBar().get_messages(Qgis.Critical)

    assert not success
    assert task_runner.fail
    assert "custom failure:" in messages


def test_function_task_without_params(task_runner: TestTaskRunner):
    task = FunctionTask(fn)
    success = task_runner.run_task(task)

    assert success
    assert task.result == ((), {})


def test_function_task_with_params(task_runner: TestTaskRunner):
    task = FunctionTask(lambda: fn(1, 2, a=1, b=2))
    success = task_runner.run_task(task)

    assert success
    assert task.result == ((1, 2), {"a": 1, "b": 2})



"""QListWidget with fields selection."""
# flake8: noqa ANN001, ANN204, ANN201
from qgis.core import QgsVectorLayer
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import QAbstractItemView, QListWidget, QListWidgetItem

__copyright__ = "Copyright 2019, 3Liz"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


class ListFieldsSelection(QListWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSelectionMode(QAbstractItemView.MultiSelection)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.layer = None

    def set_layer(self, layer: QgsVectorLayer):
        self.layer = layer

        self.clear()

        for field in self.layer.fields():
            cell = QListWidgetItem()
            alias = field.alias()
            if not alias:
                cell.setText(field.name())
            else:
                cell.setText("{} ({})".format(field.name(), alias))
            cell.setData(Qt.UserRole, field.name())
            index = layer.fields().indexFromName(field.name())
            if index >= 0:
                cell.setIcon(self.layer.fields().iconForField(index))
            self.addItem(cell)

    def set_selection(self, fields: tuple):
        for i in range(self.count()):
            item = self.item(i)
            item.setSelected(item.data(Qt.UserRole) in fields)

    def selection(self) -> list:
        selection = []
        for i in range(self.count()):
            item = self.item(i)
            if item.isSelected():
                selection.append(item.data(Qt.UserRole))
        return selection



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 2"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

from typing import Any, Callable

from .exceptions import QgsPluginException
from .i18n import tr
from .messages import MsgBar
from .tasks import FunctionTask


def log_if_fails(fn: Callable) -> Callable:
    """
    Use this as a decorator with functions and methods that
    might throw uncaught exceptions.
    """
    from functools import wraps

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> None:  # noqa: ANN001
        try:
            # Qt injects False into some signals
            if args[1:] != (False,):
                fn(*args, **kwargs)
            else:
                fn(*args[:-1], **kwargs)
        except QgsPluginException as e:
            MsgBar.exception(e, **e.bar_msg, stack_info=True)
        except Exception as e:
            MsgBar.exception(tr("Unhandled exception occurred"), e, stack_info=True)

    return wrapper


def taskify(fn: Callable) -> Callable:
    """
    Decoration used to turn any function or method into a FunctionTask task.
    """
    from functools import wraps

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> FunctionTask:  # noqa: ANN001
        return FunctionTask(lambda: fn(*args, **kwargs))

    return wrapper



"""Setting up logging using QGIS, file, Sentry..."""

import functools
import logging
from enum import Enum, unique
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from qgis.core import Qgis, QgsApplication, QgsMessageLog
from qgis.gui import QgisInterface, QgsMessageBar
from qgis.PyQt.QtCore import QObject, pyqtSignal, pyqtSlot
from qgis.PyQt.QtWidgets import QLayout, QVBoxLayout, QWidget

from .i18n import tr
from .resources import plugin_name, plugin_path, profile_path
from .settings import get_setting, setting_key

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"


@unique
class LogTarget(Enum):
    """Log target with default logging level as value"""

    STREAM = {"id": "stream", "default": "INFO"}
    FILE = {"id": "file", "default": "INFO"}
    BAR = {"id": "bar", "default": "INFO"}

    @property
    def id(self) -> str:
        return self.value["id"]

    @property
    def default_level(self) -> str:
        return self.value["default"]


def qgis_level(logging_level: str) -> int:
    """Check for the corresponding QGIS Level according to Logging Level.

    For QGIS:
    https://qgis.org/api/classQgis.html#a60c079f4d8b7c479498be3d42ec96257

    For Logging:
    https://docs.python.org/3/library/logging.html#levels

    :param logging_level: The Logging level
    :target logging_level: basestring

    :return: The QGIS Level
    :rtype: Qgis.MessageLevel
    """
    if logging_level == "CRITICAL":
        return Qgis.Critical
    elif logging_level == "ERROR":
        return Qgis.Critical
    elif logging_level == "WARNING":
        return Qgis.Warning
    elif logging_level == "INFO":
        return Qgis.Info
    elif logging_level == "DEBUG":
        return Qgis.Info

    return Qgis.Info


def bar_msg(
    details: Any = "", duration: Optional[int] = None, success: bool = False
) -> Dict[str, Any]:
    """
    Helper function to construct extra arguments for message bar logger message

    :param details: Longer body of the message. Can be set to empty string.
    :param duration: can be used to specify the message timeout in seconds. If
        ``duration`` is set to 0, then the message must be manually dismissed
        by the user.
    :param success: Whether the message is success message or not
    """
    args = {"details": str(details), "success": success}
    if duration is not None:
        args["duration"] = duration
    return args


class QgsLogHandler(logging.Handler):
    """A logging handler that will log messages to the QGIS logging console"""

    def __init__(
        self, level: int = logging.NOTSET, message_log_name: Optional[str] = None
    ) -> None:
        logging.Handler.__init__(self, level)
        self._message_log_name = message_log_name

    def emit(self, record: logging.LogRecord) -> None:
        """Try to log the message to QGIS if available, otherwise do nothing.

        :param record: logging record containing whatever info needs to be
                logged.
        """
        tag_kwargs = (
            {} if self._message_log_name is None else {"tag": self._message_log_name}
        )
        try:
            # noinspection PyCallByClass,PyTypeChecker
            QgsMessageLog.logMessage(
                record.getMessage(), level=qgis_level(record.levelname), **tag_kwargs
            )
        except MemoryError:
            message = tr(
                "Due to memory limitations on this machine, {} logger can not "
                "handle the full log"
            ).format(self._message_log_name)
            # print(message)
            # noinspection PyCallByClass,PyTypeChecker
            QgsMessageLog.logMessage(message, level=Qgis.Critical, **tag_kwargs)


class QgsMessageBarFilter(logging.Filter):
    """
    A logging filter to decide whether the message should be passed and
    to QgsMessageBarHandler as enriched

    Description of keys:
        details: Longer body of the message. Can be set to empty string.
        duration: can be used to specify the message timeout in seconds. If ``duration``
            is set to 0, then the message must be manually dismissed by the user.
        success: boolean, defaults to False.
            Whether the message is success message or not
    """

    def filter(self, record: logging.LogRecord) -> bool:
        args = record.__dict__
        if "details" not in args:
            return False

        record.qgis_level = (  # type: ignore
            qgis_level(record.levelname)
            if not args.get("success", False)
            else Qgis.Success
        )
        record.duration = args.get("duration", self.bar_msg_duration(record.levelname))  # type: ignore # noqa E501
        return True

    @staticmethod
    def bar_msg_duration(logging_level: str) -> int:
        """Check default duration for messages in message bar based on level.

        :param logging_level: The Logging level
        :return: duration in seconds
        """
        if logging_level == "CRITICAL":
            return 12
        elif logging_level == "ERROR":
            return 10
        elif logging_level == "WARNING":
            return 6
        elif logging_level == "INFO":
            return 4
        elif logging_level == "DEBUG":
            return 4

        return 4


class SimpleMessageBarProxy(QObject):
    """Signal-slot pair to always push messages in the main thread."""

    _emit_message = pyqtSignal(str, str, int, int)

    def __init__(self, msg_bar: Optional[QgsMessageBar] = None) -> None:
        super().__init__()
        self._msg_bar = msg_bar
        self._emit_message.connect(self.push_message)

    def emit_message(self, title: str, text: str, level: int, duration: int) -> None:
        self._emit_message.emit(title, text, level, duration)

    @pyqtSlot(str, str, int, int)
    def push_message(self, title: str, text: str, level: int, duration: int) -> None:
        try:
            if self._msg_bar is not None:
                self._msg_bar.pushMessage(
                    title=title,
                    text=text,
                    level=level,
                    duration=duration,
                )
        except Exception:
            pass


class QgsMessageBarHandler(logging.Handler):
    """A logging handler that will log messages to the QGIS message bar."""

    def __init__(self, msg_bar: Optional[QgsMessageBar] = None) -> None:
        super().__init__()
        self._message_bar_proxy = SimpleMessageBarProxy(msg_bar)
        self._message_bar_proxy.moveToThread(QgsApplication.instance().thread())

    def emit(self, record: logging.LogRecord) -> None:
        """
        Push info message to the QGIS message bar. Pass "extra" kwarg
        to logger to use with mandatory "details" key.

        :param record: logging record enriched with extra information from
            QgsMessageBarFilter
        """
        self._message_bar_proxy.emit_message(
            record.message,
            record.details,  # type: ignore
            record.qgis_level,  # type: ignore
            record.duration,  # type: ignore
        )


def add_logging_handler_once(logger: logging.Logger, handler: logging.Handler) -> bool:
    """A helper to add a handler to a logger, ensuring there are no duplicates.

    :param logger: Logger that should have a handler added.
    :type logger: logging.logger

    :param handler: Handler instance to be added. It will not be added if an
        instance of that Handler subclass already exists.
    :type handler: logging.Handler

    :returns: True if the logging handler was added, otherwise False.
    :rtype: bool
    """
    class_name = handler.__class__.__name__
    for logger_handler in logger.handlers:
        if logger_handler.__class__.__name__ == class_name:
            return False

    logger.addHandler(handler)
    return True


def get_log_level_key(target: LogTarget) -> str:
    """Finds QSetting key for log level"""
    return setting_key("log_level", target.id)


def get_log_level_name(target: LogTarget) -> str:
    """Finds the log level name of the target"""
    return get_setting(get_log_level_key(target), target.default_level, str)


def get_log_level(target: LogTarget) -> int:
    """Finds log level of the target"""
    return logging.getLevelName(get_log_level_name(target))


def get_log_folder() -> Path:
    """
    Get Path to the log folder in QGIS profile directory.
    If it does not exist, create one.

    :return: Path to the log folder
    """
    old_log_dir = Path(plugin_path("logs"))
    if old_log_dir.exists():
        return old_log_dir
    log_dir = Path(profile_path("logs"))
    log_dir.mkdir(exist_ok=True)
    return log_dir


def _create_handlers(
    message_log_name: str, message_bar: Optional[QgsMessageBar]
) -> List[logging.Handler]:
    handlers: List[logging.Handler] = []

    stream_level = get_log_level(LogTarget.STREAM)
    if stream_level > logging.NOTSET:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(stream_level)
        console_formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(message)s", "%d.%m.%Y %H:%M:%S"
        )
        console_handler.setFormatter(console_formatter)
        handlers.append(console_handler)

    file_level = get_log_level(LogTarget.FILE)
    if file_level > logging.NOTSET:
        log_file_name = (
            "".join((c if c.isalnum() else "_") for c in message_log_name) + ".log"
        )
        file_handler = RotatingFileHandler(
            str(get_log_folder() / log_file_name), maxBytes=1024 * 1024 * 2
        )
        file_handler.setLevel(file_level)
        file_formatter = logging.Formatter(
            "%(asctime)s - [%(levelname)-7s] - %(filename)s:%(lineno)d : %(message)s",
            "%d.%m.%Y %H:%M:%S",
        )
        file_handler.setFormatter(file_formatter)
        handlers.append(file_handler)

    bar_level = get_log_level(LogTarget.BAR)
    if bar_level > logging.NOTSET and message_bar is not None:
        qgis_msg_bar_handler = QgsMessageBarHandler(message_bar)
        qgis_msg_bar_handler.addFilter(QgsMessageBarFilter())
        qgis_msg_bar_handler.setLevel(bar_level)
        handlers.append(qgis_msg_bar_handler)

    # NOTE: NOTSET on handler will match everything, compared to NOTSET
    # on logger will match nothing. use the lowest active level from
    # the previous specific logging handlers for the qgis message log handler.
    # if nothing was enabled, use NOTSET for this handler, and later set logger
    # to use the lowest level on the handlers returned from this function
    qgis_message_log_handler = QgsLogHandler(message_log_name=message_log_name)
    qgis_message_log_handler.setLevel(
        logging.NOTSET if len(handlers) == 0 else min(h.level for h in handlers)
    )
    qgis_message_log_formatter = logging.Formatter("[%(levelname)-7s]- %(message)s")
    qgis_message_log_handler.setFormatter(qgis_message_log_formatter)
    handlers.append(qgis_message_log_handler)

    return handlers


def setup_logger(  # noqa QGS105
    logger_name: str, iface: Optional[QgisInterface] = None
) -> logging.Logger:
    """Run once when the module is loaded and enable logging.


    :param logger_name: The logger name that we want to set up.
    :param iface: QGIS Interface

    Borrowed heavily from this:
    http://docs.python.org/howto/logging-cookbook.html

    Now to log a message do::
       LOGGER.debug('Some debug message')

    And to a message bar::
       LOGGER.info('Some bar message', extra={'details': 'details'})
       LOGGER.info('Some bar message', extra=bar_msg('details')) # With helper function
    """

    if iface is None:
        try:
            from qgis.utils import iface  # type: ignore
        except ImportError:
            iface = None

    if iface is not None:
        message_bar = iface.messageBar()
    else:
        message_bar = None

    # keep api stable and create the handlers as if the passed logger name
    # was plugin_name()
    handlers = _create_handlers(plugin_name(), message_bar)

    # if the logger name was plugin_name(), create also the necessary logger for the
    # qgis_plugin_tools namespace for MsgBar and others to work properly using __name__
    logger_names = [logger_name]
    if logger_name == plugin_name():
        logger_names.append(__name__.replace(".tools.custom_logging", "", 1))

    for logger_name in logger_names:
        logger = logging.getLogger(logger_name)

        # take the lowest level from the enabled handlers
        logger.setLevel(min(h.level for h in handlers))

        for handler in handlers:
            add_logging_handler_once(logger, handler)

    return logging.getLogger(logger_name)


def add_logger_msg_bar_to_widget(logger_name: str, widget: QWidget) -> None:
    """
    Adds QgsMessageBar to any widget if it is not already there.
    This message bar will be used in logging instead of iface message bar.
    :param logger_name: The logger name that we want modify
    :param widget: QWidget that will have the message bar
    """
    if not hasattr(widget, "message_bar"):
        layout: QLayout = widget.layout()
        widget.message_bar = QgsMessageBar(widget)  # type: ignore
        if isinstance(layout, QVBoxLayout):
            # noinspection PyArgumentList
            layout.insertWidget(0, widget.message_bar)
    use_custom_msg_bar_in_logger(logger_name, widget.message_bar)


def use_custom_msg_bar_in_logger(logger_name: str, msg_bar: QgsMessageBar) -> None:
    """
    Remove QgsMessageBarHandler that is using the iface message bar
    and use custom message bar instead

    :param logger_name: The logger name that we want modify
    :param msg_bar: message bar that is inside dialog, dockwidget or
        in some other widget
    """
    logger = logging.getLogger(logger_name)
    bar_level = get_log_level(LogTarget.BAR)

    for handler in logger.handlers[:]:
        if isinstance(handler, QgsMessageBarHandler):
            logger.removeHandler(handler)

    qgis_msg_bar_handler = QgsMessageBarHandler(msg_bar)
    qgis_msg_bar_handler.addFilter(QgsMessageBarFilter())
    qgis_msg_bar_handler.setLevel(bar_level)
    add_logging_handler_once(logger, qgis_msg_bar_handler)


def teardown_logger(logger_name: str) -> None:
    """Remove all handlers from the logger

    :param logger_name: The logger name that we want to tear down.
    """
    logger = logging.getLogger(logger_name)
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # if the logger name was plugin_name(), also clean up the special case logger
    if logger_name == plugin_name():
        teardown_logger(__name__.replace(".tools.custom_logging", "", 1))


def teardown_loggers(logger_names: List[str]) -> None:
    """
    Remove the added handlers from the speficied handler.
    """
    for logger_name in logger_names:
        teardown_logger(logger_name)


def setup_loggers(
    *logger_names: str,
    message_log_name: str,
    message_bar: Optional[QgsMessageBar] = None,
) -> Callable[[], None]:
    """
    Setups all the loggers for the given logger names.

    Returns a teardown callback so setup can be called in initGui and
    the returned callback in unload.
    """
    if message_bar is None:
        try:
            from qgis.utils import iface

            message_bar = iface.messageBar()
        except ImportError:
            message_bar = None

    handlers = _create_handlers(message_log_name, message_bar)

    for logger_name in logger_names:
        logger = logging.getLogger(logger_name)

        # take the lowest level from the enabled handlers
        logger.setLevel(min(h.level for h in handlers))

        for handler in handlers:
            add_logging_handler_once(logger, handler)

    return functools.partial(teardown_loggers, logger_names)



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

from qgis.core import QgsWkbTypes

from ..tools.layers import LayerType


def test_layer_type():
    assert LayerType.from_wkb_type(QgsWkbTypes.CurvePolygonZM) == LayerType.Polygon
    assert LayerType.from_wkb_type(QgsWkbTypes.MultiPoint) == LayerType.Point
    assert LayerType.from_wkb_type(QgsWkbTypes.MultiLineString25D) == LayerType.Line



"""I18n tools."""

from os.path import join
from typing import Any, Optional, Tuple

from qgis.core import QgsSettings
from qgis.PyQt.QtCore import QFileInfo, QLocale
from qgis.PyQt.QtWidgets import QApplication

from .resources import plugin_name, plugin_path, resources_path, slug_name

__copyright__ = "Copyright 2019, 3Liz, 2020-2021 Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


def setup_translation(
    file_pattern: str = "{}.qm", folder: Optional[str] = None
) -> Tuple[str, Optional[str]]:
    """Find the translation file according to locale.

    :param file_pattern: Custom file pattern to use to find QM files.
    :type file_pattern: basestring

    :param folder: Optional folder to look in if it's not the default.
    :type folder: basestring

    :return: The locale and the file path to the QM file, or None.
    :rtype: (basestring, basestring)
    """
    locale = QgsSettings().value("locale/userLocale", QLocale().name())

    for prefix in ["", f"{plugin_name()}_", f"{slug_name()}_"]:
        for fldr in [folder, plugin_path("i18n"), resources_path("i18n")]:
            prefixed_locale = prefix + locale
            if fldr:
                ts_file = QFileInfo(join(fldr, file_pattern.format(prefixed_locale)))
                if ts_file.exists():
                    return locale, ts_file.absoluteFilePath()

            prefixed_locale = prefix + locale[0:2]
            if fldr:
                ts_file = QFileInfo(join(fldr, file_pattern.format(prefixed_locale)))
                if ts_file.exists():
                    return locale, ts_file.absoluteFilePath()

    return locale, None


def tr(
    text: str,
    *args: Any,
    context: str = "@default",
    **kwargs: Any,
) -> str:
    """Get the translation for a string using Qt translation API.

    We implement this ourselves since we do not inherit QObject.

    :param text: String for translation.
    :param args: arguments to use in formatting.
    :param context: Context of the translation.
    :param kwargs: keyword arguments to use in formatting.

    :returns: Translated version of message.
    """
    # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
    return QApplication.translate(context, text).format(*args, **kwargs)



"""Tools about version."""
from typing import Tuple

from osgeo import osr

from .exceptions import QgsPluginVersionInInvalidFormat
from .resources import metadata_config


def format_version_integer(version_string: str) -> int:
    """Transform version string to integers to allow comparing versions.

    Transform "0.1.2" into "000102"
    Transform "10.9.12" into "100912"
    """
    return int("".join([a.zfill(2) for a in version_string.strip().split(".")]))


def version(remove_v_prefix: bool = True) -> str:
    """Return the version defined in metadata.txt."""
    v = metadata_config()["general"]["version"]
    if v.startswith("v") and remove_v_prefix:
        v = v[1:]
    return v


def proj_version() -> Tuple[int, int]:
    """Returns PROJ library version"""
    major: int = osr.GetPROJVersionMajor()
    minor: int = osr.GetPROJVersionMinor()
    return major, minor


def version_from_string(version: str) -> Tuple[int, int, int]:
    """
    Transforms version string in format 'x.y.z' to tuple (x,y,z) for comparisons
    :param version:
    :return:
    """
    parts = version.split(".")
    if len(parts) != 3:
        raise QgsPluginVersionInInvalidFormat()
    return int(parts[0]), int(parts[1]), int(parts[2])


def string_from_version(version: Tuple[int, int, int]) -> str:
    """
    Transforms version tuple in format (x,y,z) to string in format 'x.y.z'
    :param version:
    :return:
    """
    if len(version) != 3:
        raise QgsPluginVersionInInvalidFormat()
    return ".".join(map(str, version))



"""QListWidget with layers selection."""
from typing import Optional

from qgis.core import QgsMapLayer, QgsMapLayerModel, QgsProject
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import QAbstractItemView, QListWidget, QListWidgetItem, QWidget

__copyright__ = "Copyright 2019, 3Liz"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


class ListLayersSelection(QListWidget):
    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.setSelectionMode(QAbstractItemView.MultiSelection)
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.project: Optional[QgsProject] = None

    def set_project(self, project: QgsProject) -> None:
        self.project = project

        self.clear()

        for layer in self.project.mapLayers().values():
            if layer.type() != QgsMapLayer.VectorLayer:
                continue

            if not layer.isSpatial():
                continue

            cell = QListWidgetItem()
            cell.setText(layer.name())
            cell.setData(Qt.UserRole, layer.id())
            cell.setIcon(QgsMapLayerModel.iconForLayer(layer))
            self.addItem(cell)

    def set_selection(self, layers: tuple) -> None:
        for i in range(self.count()):
            item = self.item(i)
            item.setSelected(item.data(Qt.UserRole) in layers)

    def selection(self) -> list:
        selection = []
        for i in range(self.count()):
            item = self.item(i)
            if item.isSelected():
                selection.append(item.data(Qt.UserRole))
        return selection



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import pytest

from ..tools.exceptions import QgsPluginInvalidProjectSetting
from ..tools.settings import (
    get_project_setting,
    get_setting,
    set_project_setting,
    set_setting,
)


def test_set_setting(qgis_new_project):
    set_setting("test_setting", "test_value")
    assert get_setting("test_setting") == "test_value"


def test_get_setting(qgis_new_project):
    assert get_setting("non-existent", 2, int) == 2


def test_get_setting2(qgis_new_project):
    assert get_setting("non-existent", 2, str) == "2"


def test_get_setting3(qgis_new_project):
    assert get_setting("non-existent", 2, bool) is True


def test_set_project_setting(qgis_new_project):
    set_project_setting("test_setting", "test_value")
    assert get_project_setting("test_setting") == "test_value"


def test_get_project_setting(qgis_new_project):
    assert get_project_setting("non-existent", 2, int) == 2


def test_get_project_setting2(qgis_new_project):
    assert get_project_setting("non-existent", "2", str) == "2"


def test_get_project_setting3(qgis_new_project):
    assert get_project_setting("non-existent", True, bool) is True


def test_get_project_setting_throws_error(qgis_new_project):
    with pytest.raises(QgsPluginInvalidProjectSetting):
        get_project_setting("non-existent", 2, list)



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import datetime

import pytest
from qgis.core import QgsDateTimeRange, QgsRasterLayer, QgsSingleBandGrayRenderer
from qgis.PyQt.QtCore import QDateTime, Qt

from ..testing.utilities import qgis_supports_temporal
from ..tools.network import download_to_file
from ..tools.raster_layers import (
    set_band_based_on_range,
    set_fixed_temporal_range,
    set_raster_renderer_to_singleband,
)


@pytest.fixture
def netcdf_layer(tmpdir) -> QgsRasterLayer:
    path = download_to_file(
        "https://raw.githubusercontent.com/GispoCoding/FMI2QGIS/master/FMI2QGIS/test/data/aq_small.nc",  # noqa E501
        tmpdir,
    )
    return QgsRasterLayer(str(path))


@pytest.fixture
def t_range() -> QgsDateTimeRange:
    return QgsDateTimeRange(
        QDateTime(2020, 11, 2, 15, 0, 0, 0), QDateTime(2020, 11, 3, 11, 0, 0, 0)
    )


@pytest.fixture
def configured_layer(netcdf_layer, t_range) -> QgsRasterLayer:
    set_raster_renderer_to_singleband(netcdf_layer)
    set_fixed_temporal_range(netcdf_layer, t_range)
    return netcdf_layer


def test_set_raster_renderer_to_singleband(netcdf_layer):
    set_raster_renderer_to_singleband(netcdf_layer)
    assert isinstance(netcdf_layer.renderer(), QgsSingleBandGrayRenderer)


@pytest.mark.skipif(
    not qgis_supports_temporal(), reason="QGIS version does not support temporal utils"
)
def test_set_fixed_temporal_range(netcdf_layer, t_range):
    set_fixed_temporal_range(netcdf_layer, t_range)
    tprops = netcdf_layer.temporalProperties()
    assert tprops.isActive()
    assert tprops.fixedTemporalRange() == QgsDateTimeRange(
        QDateTime(2020, 11, 2, 15, 0, 0, 0, Qt.TimeSpec(1)),
        QDateTime(2020, 11, 3, 11, 0, 0, 0, Qt.TimeSpec(1)),
    )


@pytest.mark.skipif(
    not qgis_supports_temporal(), reason="QGIS version does not support temporal utils"
)
def test_set_band_based_on_range(configured_layer):
    t_range2 = QgsDateTimeRange(
        datetime.datetime(2020, 11, 2, 15, 0), datetime.datetime(2020, 11, 2, 16, 0)
    )
    band_number = set_band_based_on_range(configured_layer, t_range2)
    assert band_number == 1


@pytest.mark.skipif(
    not qgis_supports_temporal(), reason="QGIS version does not support temporal utils"
)
def test_set_band_based_on_range2(configured_layer):
    t_range2 = QgsDateTimeRange(
        datetime.datetime(2020, 11, 2, 17, 0), datetime.datetime(2020, 11, 2, 18, 0)
    )
    band_number = set_band_based_on_range(configured_layer, t_range2)
    assert band_number == 3


@pytest.mark.skipif(
    not qgis_supports_temporal(), reason="QGIS version does not support temporal utils"
)
def test_set_band_based_on_range3(configured_layer):
    t_range2 = QgsDateTimeRange(
        datetime.datetime(2020, 11, 2, 18, 0), datetime.datetime(2020, 11, 2, 22, 0)
    )
    band_number = set_band_based_on_range(configured_layer, t_range2)
    assert band_number == 4


@pytest.mark.skipif(
    not qgis_supports_temporal(), reason="QGIS version does not support temporal utils"
)
def test_set_band_based_on_range4(configured_layer):
    t_range2 = QgsDateTimeRange(
        QDateTime(2020, 11, 3, 10, 0, 0, 0, Qt.TimeSpec(1)),
        QDateTime(2020, 11, 3, 11, 0, 0, 0, Qt.TimeSpec(1)),
    )
    band_number = set_band_based_on_range(configured_layer, t_range2)
    assert band_number == 20



import json
import logging
import re
import shutil
from pathlib import Path
from typing import Dict, List, Literal, NamedTuple, Optional, Tuple
from urllib.parse import urlencode
from uuid import uuid4

from qgis.core import Qgis, QgsBlockingNetworkRequest, QgsNetworkReplyContent
from qgis.PyQt.QtCore import QByteArray, QSettings, QUrl
from qgis.PyQt.QtNetwork import QNetworkReply, QNetworkRequest

from ..tools.exceptions import QgsPluginNetworkException
from ..tools.i18n import tr
from ..tools.resources import plugin_name
from .custom_logging import bar_msg

try:
    import requests
    from requests.exceptions import RequestException
except ImportError:
    requests = None  # type: ignore
    RequestException = None  # type: ignore

__copyright__ = "Copyright 2020-2023, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

LOGGER = logging.getLogger(__name__)
ENCODING = "utf-8"
CONTENT_DISPOSITION_HEADER = "Content-Disposition"
CONTENT_DISPOSITION_BYTE_HEADER = QByteArray(
    bytes(CONTENT_DISPOSITION_HEADER, ENCODING)
)


class FileInfo(NamedTuple):
    file_name: str
    content: bytes
    content_type: str


class FileField(NamedTuple):
    field_name: str
    file_info: FileInfo


def fetch(
    url: str,
    encoding: str = ENCODING,
    authcfg_id: str = "",
    params: Optional[Dict[str, str]] = None,
) -> str:
    """
    Fetch resource from the internet. Similar to requests.get(url) but is
    recommended way of handling requests in QGIS plugin
    :param url: address of the web resource
    :param encoding: Encoding which will be used to decode the bytes
    :param authcfg_id: authcfg id from QGIS settings, defaults to ''
    :param params: Dictionary to send in the query string
    :return: encoded string of the content
    """
    content, _ = fetch_raw(url, encoding, authcfg_id, params)
    return content.decode(ENCODING)


def post(
    url: str,
    encoding: str = ENCODING,
    authcfg_id: str = "",
    data: Optional[Dict[str, str]] = None,
    files: Optional[List[FileField]] = None,
) -> str:
    """
    Post resource. Similar to requests.post(url, data, files) but is
    recommended way of handling requests in QGIS plugin
    :param url: address of the web resource
    :param encoding: Encoding which will be used to decode the bytes
    :param authcfg_id: authcfg id from QGIS settings, defaults to ''
    :param data: Dictionary to send in the request body
    :param files: Files to send multipart-encoded. Same format as requests.
    :return: encoded string of the content
    """
    content, _ = post_raw(url, encoding, authcfg_id, data, files)
    return content.decode(ENCODING)


def fetch_raw(
    url: str,
    encoding: str = ENCODING,
    authcfg_id: str = "",
    params: Optional[Dict[str, str]] = None,
) -> Tuple[bytes, str]:
    """
    Fetch resource from the internet. Similar to requests.get(url) but is
    recommended way of handling requests in QGIS plugin
    :param url: address of the web resource
    :param encoding: Encoding which will be used to decode the bytes
    :param authcfg_id: authcfg id from QGIS settings, defaults to ''
    :param params: Dictionary to send in the query string
    :return: bytes of the content and default name of the file or empty string
    """
    return request_raw(url, "get", encoding, authcfg_id, params)


def post_raw(
    url: str,
    encoding: str = ENCODING,
    authcfg_id: str = "",
    data: Optional[Dict[str, str]] = None,
    files: Optional[List[FileField]] = None,
) -> Tuple[bytes, str]:
    """
    Post resource. Similar to requests.post(url, data, files) but is
    recommended way of handling requests in QGIS plugin
    :param url: address of the web resource
    :param encoding: Encoding which will be used to decode the bytes
    :param authcfg_id: authcfg id from QGIS settings, defaults to ''
    :param data: Dictionary to send in the request body
    :param files: Files to send multipart-encoded. Same format as requests.
    :return: bytes of the content and default name of the file or empty string
    """
    return request_raw(url, "post", encoding, authcfg_id, None, data, files)


def request_raw(
    url: str,
    method: Literal["get", "post"] = "get",
    encoding: str = ENCODING,
    authcfg_id: str = "",
    params: Optional[Dict[str, str]] = None,
    data: Optional[Dict[str, str]] = None,
    files: Optional[List[FileField]] = None,
) -> Tuple[bytes, str]:
    """
    Request resource from the internet. Similar to requests.get(url) and
    requests.post(url, data) but is recommended way of handling requests in QGIS plugin
    :param url: address of the web resource
    :param method: method to use, defaults to 'get'
    :param encoding: Encoding which will be used to decode the bytes
    :param authcfg_id: authcfg id from QGIS settings, defaults to ''
    :param params: Dictionary to send in the query string
    :param data: Dictionary to send in the request body
    :param files: Files to send multipart-encoded. Same format as requests.
    :return: bytes of the content and default name of the file or empty string
    """
    if params:
        url += "?" + urlencode(params)
    LOGGER.debug(url)
    req = QNetworkRequest(QUrl(url))
    # http://osgeo-org.1560.x6.nabble.com/QGIS-Developer-Do-we-have-a-User-Agent-string-for-QGIS-td5360740.html
    user_agent = QSettings().value("/qgis/networkAndProxy/userAgent", "Mozilla/5.0")
    user_agent += " " if len(user_agent) else ""
    # noinspection PyUnresolvedReferences
    user_agent += f"QGIS/{Qgis.QGIS_VERSION_INT}"
    user_agent += f" {plugin_name()}"
    # https://www.riverbankcomputing.com/pipermail/pyqt/2016-May/037514.html
    req.setRawHeader(b"User-Agent", bytes(user_agent, encoding))
    request_blocking = QgsBlockingNetworkRequest()
    if authcfg_id:
        request_blocking.setAuthCfg(authcfg_id)
    # QNetworkRequest *only* supports get and post. No idea why.
    if method == "get":
        _ = request_blocking.get(req)
    elif method == "post":
        if data:
            # Support JSON
            byte_data = bytes(json.dumps(data), encoding)
            req.setRawHeader(
                b"Content-Type",
                bytes(f"application/json; charset={encoding}", encoding),
            )
        elif files:
            # Support multipart binary. Generate boundary like
            # https://github.com/requests/toolbelt/blob/master/requests_toolbelt/multipart/encoder.py
            boundary = uuid4().hex
            byte_boundary = bytes(f"\r\n--{boundary}\r\n", encoding)
            last_byte_boundary = bytes(f"\r\n--{boundary}--\r\n", encoding)

            # each file may have different content type, name and filename
            byte_data = b""
            for file_field in files:
                name = file_field[0]
                file_info = file_field[1]
                file_name = file_info[0]
                content = file_info[1]
                content_type = file_info[2]
                content_disposition_form_data = (
                    f"Content-Disposition: form-data;"
                    f' name="{name}";'
                    f' filename="{file_name}"\r\n'
                )
                content_type_form_data = f"Content-Type: {content_type}\r\n\r\n"
                byte_boundary_with_headers = (
                    byte_boundary
                    + bytes(content_disposition_form_data, encoding)
                    + bytes(content_type_form_data, encoding)
                )
                byte_data += byte_boundary_with_headers + content
            byte_data += last_byte_boundary
            req.setRawHeader(
                b"Content-Type",
                bytes(f"multipart/form-data; boundary={boundary}", encoding),
            )
        else:
            byte_data = b""
        _ = request_blocking.post(req, byte_data)
    else:
        raise Exception(f"Request method {method} not supported.")
    reply: QgsNetworkReplyContent = request_blocking.reply()
    reply_error = reply.error()
    if reply_error != QNetworkReply.NoError:
        # Error content will be empty in older QGIS versions:
        # https://github.com/qgis/QGIS/issues/42442
        message = (
            bytes(reply.content()).decode("utf-8")
            if len(bytes(reply.content()))
            else None
        )
        # bar_msg will just show a generic Qt error string.
        raise QgsPluginNetworkException(
            message=message,
            error=reply_error,
            bar_msg=bar_msg(reply.errorString()),
        )

    # https://stackoverflow.com/a/39103880/10068922
    default_name = ""
    if reply.hasRawHeader(CONTENT_DISPOSITION_BYTE_HEADER):
        header: QByteArray = reply.rawHeader(CONTENT_DISPOSITION_BYTE_HEADER)
        default_name = bytes(header).decode(encoding).split("filename=")[1]
        if default_name[0] in ['"', "'"]:
            default_name = default_name[1:-1]

    return bytes(reply.content()), default_name


def download_to_file(
    url: str,
    output_dir: Path,
    output_name: Optional[str] = None,
    use_requests_if_available: bool = True,
    encoding: str = ENCODING,
) -> Path:
    """
    Downloads a binary file to the file efficiently
    :param url: Url of the file
    :param output_dir: Path to the output directory
    :param output_name: If given, use this as file name. Otherwise reads file name from
    Content-Disposition header or uses the url
    :param use_requests_if_available: Use Python package requests
    if it is available in the environment
    :param encoding: Encoding which will be used to decode the bytes
    :return: Path to the file
    """

    def get_output(default_filename: str) -> Path:
        if output_name is None:
            if default_filename != "":
                out_name = default_filename
            else:
                out_name = url.replace("http://", "").replace("https://", "")
                if len(out_name.split("/")[-1]) > 2:
                    out_name = out_name.split("/")[-1]
        else:
            out_name = output_name
        return Path(output_dir, out_name)

    output = Path()

    if use_requests_if_available and requests is not None:
        # https://stackoverflow.com/a/39217788/10068922

        try:
            with requests.get(url, stream=True) as r:
                try:
                    r.raise_for_status()
                except Exception:
                    raise QgsPluginNetworkException(
                        tr("Request failed with status code {}", r.status_code),
                        bar_msg=bar_msg(r.text),
                    )
                default_filenames = re.findall(
                    "filename=(.+)", r.headers.get(CONTENT_DISPOSITION_HEADER, "")
                )
                default_filename = (
                    default_filenames[0] if len(default_filenames) else ""
                )
                output = get_output(default_filename)
                with open(output, "wb") as f:
                    shutil.copyfileobj(r.raw, f)
        except RequestException as e:
            raise QgsPluginNetworkException(tr("Request failed"), bar_msg=bar_msg(e))
    else:
        # Using simple fetch_raw
        content, default_filename = fetch_raw(url, encoding)
        output = get_output(default_filename)
        with open(output, "wb") as f:
            f.write(content)
    return output



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import json

import pytest

from ..tools.exceptions import QgsPluginNetworkException
from ..tools.network import download_to_file, fetch, post


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_fetch(qgis_new_project):
    data = fetch("https://httpbin.org/get")
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/get"


def test_fetch_invalid_url(qgis_new_project):
    with pytest.raises(QgsPluginNetworkException):
        fetch("invalidurl")


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_fetch_params(qgis_new_project):
    data = fetch("https://httpbin.org/get", params={"foo": "bar"})
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/get?foo=bar"
    assert data["args"] == {"foo": "bar"}


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_post(qgis_new_project):
    data = post("https://httpbin.org/post")
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/post"


def test_post_invalid_url(qgis_new_project):
    with pytest.raises(QgsPluginNetworkException):
        post("invalidurl")


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_post_data(qgis_new_project):
    data = post("https://httpbin.org/post", data={"foo": "bar"})
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/post"
    assert data["data"] == json.dumps({"foo": "bar"})


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_upload_file(qgis_new_project, file_fixture):
    file_name, file_content, file_type = file_fixture
    data = post(
        "https://httpbin.org/post",
        files=[("file", (file_name, file_content, file_type))],
    )
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/post"
    assert data["files"]
    assert bytes(data["files"]["file"], "utf-8") == file_content


@pytest.mark.skip(
    "Skipped due httbin.org being flaky."
    "TODO: refactor tests not to be dependent of an external http api"
)
def test_upload_multiple_files(qgis_new_project, file_fixture, another_file_fixture):
    file_name, file_content, file_type = file_fixture
    another_file_name, another_file_content, another_file_type = another_file_fixture
    data = post(
        "https://httpbin.org/post",
        files=[
            ("file", (file_name, file_content, file_type)),
            (
                "another_file",
                (another_file_name, another_file_content, another_file_type),
            ),
        ],
    )
    data = json.loads(data)
    assert data["url"] == "https://httpbin.org/post"
    assert data["files"]
    assert bytes(data["files"]["file"], "utf-8") == file_content
    assert bytes(data["files"]["another_file"], "utf-8") == another_file_content


@pytest.mark.skip(
    "file does not exist. "
    "TODO: search another file to be used using Content-Disposition"
)
def test_download_to_file(qgis_new_project, tmpdir):
    path_to_file = download_to_file(
        "https://twitter.com/gispofinland/status/1324599933337567232/photo/1",
        tmpdir,
        "test_file",
    )
    assert path_to_file.exists()
    assert path_to_file.is_file()


@pytest.mark.skip(
    "file does not exist. "
    "TODO: search another file to be used using Content-Disposition"
)
def test_download_to_file_without_requests(qgis_new_project, tmpdir):
    path_to_file = download_to_file(
        "https://twitter.com/gispofinland/status/1324599933337567232/photo/1",
        tmpdir,
        "test_file",
        use_requests_if_available=False,
    )
    assert path_to_file.exists()
    assert path_to_file.is_file()


def test_download_to_file_with_name(qgis_new_project, tmpdir):
    path_to_file = download_to_file(
        "https://raw.githubusercontent.com/GispoCoding/FMI2QGIS/master/FMI2QGIS/test/data/aq_small.nc",  # noqa E501
        tmpdir,
    )
    assert path_to_file.exists()
    assert path_to_file.is_file()
    assert path_to_file.name == "aq_small.nc"


def test_download_to_file_invalid_url(qgis_new_project, tmpdir):
    with pytest.raises(QgsPluginNetworkException):
        download_to_file("invalidurl", tmpdir)


def test_download_to_file_invalid_url_without_requests(qgis_new_project, tmpdir):
    with pytest.raises(QgsPluginNetworkException):
        download_to_file("invalidurl", tmpdir)



__copyright__ = "Copyright 2021, qgis_plugin_tools contributors"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"

import logging
from typing import Callable, Optional, Union

from qgis.core import QgsApplication, QgsTask
from qgis.PyQt import uic
from qgis.PyQt.QtCore import pyqtSignal
from qgis.PyQt.QtGui import QCloseEvent
from qgis.PyQt.QtWidgets import (
    QDialog,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)

from ..tools.decorations import log_if_fails
from ..tools.i18n import tr
from ..tools.resources import qgis_plugin_tools_resources

FORM_CLASS: QWidget
FORM_CLASS, _ = uic.loadUiType(qgis_plugin_tools_resources("ui", "progress_dialog.ui"))
LOGGER = logging.getLogger(__name__)


class ProgressDialog(QDialog, FORM_CLASS):
    """
    Dialog containing progress bar to show processes of long running tasks.
    """

    progress_bar: QProgressBar
    status_label: QLabel
    v_layout: QVBoxLayout
    abort_btn_text: str = tr("Abort")

    aborted = pyqtSignal()

    def __init__(
        self,
        parent: Optional[QDialog] = None,
        show_abort_button: bool = False,
        abort_btn_text: str = abort_btn_text,
    ) -> None:
        QDialog.__init__(self, parent)
        self.setupUi(self)
        if show_abort_button:
            self.push_btn = QPushButton()
            self.push_btn.setText(abort_btn_text)
            layout = QHBoxLayout()
            spacer = QSpacerItem(
                100, 20, QSizePolicy.MinimumExpanding, QSizePolicy.Expanding
            )
            layout.addSpacerItem(spacer)
            layout.addWidget(self.push_btn)
            self.v_layout.addLayout(layout)
            self.push_btn.clicked.connect(self._aborted)

        self.update_progress_bar(0)

    def set_status(self, status_text: str) -> None:
        LOGGER.debug(f"Status:   {status_text}")
        self.status_label.setText(status_text)

    def update_progress_bar(self, progress: Union[int, float]) -> None:
        """Update progress bar with a progress"""
        self.progress_bar.setValue(int(min(100, progress)))

    def closeEvent(self, close_event: QCloseEvent) -> None:  # noqa: N802
        super().closeEvent(close_event)
        LOGGER.warning("Closing progress bar")
        self.aborted.emit()

    @log_if_fails
    def _aborted(self) -> None:
        LOGGER.warning("Aborted")
        self.close()


def create_simple_continuous_progress_dialog(
    status_text: str,
    parent: Optional[QDialog] = None,
    show_abort_button: bool = False,
    abort_btn_text: str = ProgressDialog.abort_btn_text,
) -> ProgressDialog:
    """
    Creates simple progress dialog with a continuous progress bar.
    """
    progress_dialog = ProgressDialog(parent, show_abort_button, abort_btn_text)
    progress_dialog.progress_bar.setMaximum(0)
    progress_dialog.progress_bar.setMinimum(0)
    progress_dialog.set_status(status_text)
    return progress_dialog


def run_task_with_progress_dialog(
    task: QgsTask,
    status_text: str,
    parent: Optional[QDialog] = None,
    show_abort_button: bool = False,
    abort_btn_text: str = ProgressDialog.abort_btn_text,
    completed_callback: Optional[Callable] = None,
    terminated_callback: Optional[Callable] = None,
) -> None:
    """
    Runs a given task while showing a progress bar dialog.
    """
    progress_dialog = ProgressDialog(parent, show_abort_button, abort_btn_text)
    progress_dialog.set_status(status_text)
    task.progressChanged.connect(progress_dialog.update_progress_bar)
    _make_connections_and_run_task(
        progress_dialog, task, completed_callback, terminated_callback
    )


def run_task_with_continuous_progress_dialog(
    task: QgsTask,
    status_text: str,
    parent: Optional[QDialog] = None,
    show_abort_button: bool = False,
    abort_btn_text: str = ProgressDialog.abort_btn_text,
    completed_callback: Optional[Callable] = None,
    terminated_callback: Optional[Callable] = None,
) -> None:
    """
    Runs a given task while showing a simple continuous progress bar dialog.
    """
    progress_dialog = create_simple_continuous_progress_dialog(
        status_text, parent, show_abort_button, abort_btn_text
    )
    _make_connections_and_run_task(
        progress_dialog, task, completed_callback, terminated_callback
    )


def _make_connections_and_run_task(
    progress_dialog: ProgressDialog,
    task: QgsTask,
    completed_callback: Optional[Callable],
    terminated_callback: Optional[Callable],
) -> None:
    task.taskCompleted.connect(lambda: progress_dialog.close())
    task.taskTerminated.connect(lambda: progress_dialog.close())
    progress_dialog.aborted.connect(task.cancel)
    if completed_callback:
        task.taskCompleted.connect(completed_callback)
    if terminated_callback:
        task.taskTerminated.connect(terminated_callback)
    task_manager = QgsApplication.taskManager()
    task_manager.addTask(task)
    # Wait until task is either completed or terminated
    progress_dialog.exec()



"""Tools for layers."""
from typing import List

from qgis.core import QgsLayerTreeUtils, QgsMapLayer, QgsProject

__copyright__ = "Copyright 2019, 3Liz, 2020-2021 Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


def is_ghost_layer(layer: QgsMapLayer) -> bool:
    """Check if the given layer is a ghost layer or not.

    A ghost layer is included in the QgsProject list of layers
    but is not present in the layer tree.

    :param layer: QgsMapLayer coming from the QgsProject.mapLayers.
    :type layer: QgsMapLayer

    :return: If the layer is ghost, IE not found in the legend layer tree.
    :rtype: bool
    """
    # noinspection PyArgumentList
    project = QgsProject.instance()
    # noinspection PyArgumentList
    count = QgsLayerTreeUtils.countMapLayerInTree(project.layerTreeRoot(), layer)
    return count == 0


def remove_all_ghost_layers() -> List[QgsMapLayer]:
    """Remove all ghost layers from project.

    :return: The list of layers name which have been removed.
    :rtype: list
    """
    # noinspection PyArgumentList
    project = QgsProject.instance()
    ghosts = []
    for layer in project.mapLayers().values():
        if is_ghost_layer(layer):
            ghosts.append(layer.name())
            project.removeMapLayer(layer.id())

    project.setDirty(True)
    return ghosts



# flake8: noqa N802
"""Base class algorithm."""
from os.path import isfile

from qgis.core import QgsProcessingAlgorithm
from qgis.PyQt.QtGui import QIcon

from .resources import resources_path

__copyright__ = "Copyright 2020, 3Liz"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


class BaseProcessingAlgorithm(QgsProcessingAlgorithm):
    def __init__(self):
        super().__init__()

    def createInstance(self):
        return type(self)()

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.FlagHideFromModeler

    def icon(self):
        icon = resources_path("icons", "icon.png")
        if isfile(icon):
            return QIcon(icon)
        else:
            return super().icon()

    def shortHelpString(self):
        raise NotImplementedError



from qgis.PyQt.QtWidgets import QGridLayout, QLayoutItem

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

""" Removing functionality adapted from C example
https://stackoverflow.com/a/19256990/10068922 """


def remove_row(layout: QGridLayout, row: int, delete_widgets: bool = True) -> None:
    """
    Removes the contents of the given layout row.

    :param layout:
    :param row:
    :param delete_widgets:
    :return:
    """
    _remove(layout, row, -1, delete_widgets)
    layout.setRowMinimumHeight(row, 0)
    layout.setRowStretch(row, 0)


def remove_column(
    layout: QGridLayout, column: int, delete_widgets: bool = True
) -> None:
    """
    Removes the contents of the given layout column.

    :param layout:
    :param column:
    :param delete_widgets:
    :return:
    """
    _remove(layout, -1, column, delete_widgets)
    layout.setColumnMinimumWidth(column, 0)
    layout.setColumnStretch(column, 0)


def remove_cell(
    layout: QGridLayout, row: int, column: int, delete_widgets: bool = True
) -> None:
    """
    Removes the contents of the given layout cell.

    :param layout:
    :param row:
    :param column:
    :param delete_widgets:
    :return:
    """
    _remove(layout, row, column, delete_widgets)


def _remove(layout: QGridLayout, row: int, column: int, delete_widgets: bool) -> None:
    for i in reversed(range(1, layout.count())):
        r, c, rs, cs = layout.getItemPosition(i)
        if (row == -1 or (r <= row < r + rs)) and (
            column == -1 or (c <= column < c + cs)
        ):
            item: QLayoutItem = layout.takeAt(i)

            layout.removeItem(item)
            widget = item.widget()
            if widget is not None:
                layout.removeWidget(widget)
                widget = None
            if delete_widgets:
                _delete_child_widgets(item)
            del item


def _delete_child_widgets(item: QLayoutItem) -> None:
    """
    Deletes all child widgets of the given layout item.

    :param item:
    :return:
    """
    layout = item.layout()
    if layout:
        for i in range(layout.count()):
            _delete_child_widgets(layout.itemAt(i))
        layout.deleteLater()
    widget = item.widget()
    if widget is not None:
        widget.hide()
        # noinspection PyTypeChecker
        widget.setParent(None)
        widget = None
    del item
    item = None  # noqa F841



from threading import Thread
from unittest.mock import MagicMock

from qgis.PyQt.QtCore import QCoreApplication

from ..tools.custom_logging import SimpleMessageBarProxy


def test_message_log_proxies_between_threads():
    mock_msg_bar = MagicMock()
    proxy = SimpleMessageBarProxy(mock_msg_bar)

    def mock_thread():
        proxy.emit_message("title", "text", 1, 2)

    thread = Thread(target=mock_thread)
    thread.start()

    mock_msg_bar.pushMessage.assert_not_called()
    thread.join(5)
    assert not thread.is_alive()

    QCoreApplication.processEvents()
    mock_msg_bar.pushMessage.assert_called_once_with(
        title="title", text="text", level=1, duration=2
    )



import logging
from typing import Any, Callable, Optional, Union

from qgis.core import QgsTask

from .exceptions import QgsPluginException, TaskInterruptedException
from .i18n import tr
from .messages import MsgBar

LOGGER = logging.getLogger(__name__)

__copyright__ = "Copyright 2021, qgis_plugin_tools contributors"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"


class BaseTask(QgsTask):
    """
    Base class for writing QgsTask classes.
    Provides some utility functionality, like error handling and
    easy canceling.
    """

    def __init__(self) -> None:
        super().__init__(self.name, QgsTask.CanCancel)
        self.exception: Optional[Exception] = None

    @property
    def name(self) -> str:
        return self.__class__.__name__

    def run(self) -> bool:
        """
        Run the task.

        :return: whether task finished successfully or not.
        """

        LOGGER.debug(f"Started task {self.name}")
        try:
            self._check_if_canceled()
            return self._run()
        except Exception as e:  # noqa: PIE786
            self.exception = e
            return False

    def finished(self, result: bool) -> None:
        """
        This function is automatically called when the task has completed
        (successfully or not).

        finished is always called from the main thread, so it's safe
        to do GUI operations and raise Python exceptions here.
        :param result: the return value from self.run
        """
        if result:
            LOGGER.debug(
                f"Task {self.name} ended successfully in "
                f"{self.elapsedTime() / 1000:.2f}!"
            )
            pass
        else:
            if self.exception is None:
                MsgBar.warning(
                    tr("Task {} was not successful", self.name),
                    tr("Task was cancelled by user or some dependency tasks failed"),
                )
            else:
                try:
                    raise self.exception
                except QgsPluginException as e:
                    MsgBar.exception(str(e), **e.bar_msg)
                except Exception as e:
                    MsgBar.exception(tr("Unhandled exception occurred"), e)

    def setProgress(self, progress: Union[int, float]) -> None:  # noqa: N802
        self._check_if_canceled()
        super().setProgress(progress)

    def _run(self) -> bool:
        """
        Common pitfalls:

        - Do not create and add layers to the project!

        - Do not connect or emit any PyQtSignals

        - Do not use print function!

        - Do not do anything related to GUI!
        """
        raise NotImplementedError()

    def _check_if_canceled(self) -> None:
        """Check if the task has been canceled"""
        if self.isCanceled():
            raise TaskInterruptedException(tr("Task canceled!"))


class FunctionTask(BaseTask):
    """
    Utility class for creating a task out of a function.
    """

    def __init__(self, callback_function: Callable) -> None:
        super().__init__()
        self._callback_function = callback_function
        self.result: Any = None

    def _run(self) -> bool:
        self.result = self._callback_function()
        return True



__copyright__ = "Copyright 2020, 3Liz"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"

from qgis.PyQt import Qsci
from qgis.PyQt.QtWidgets import QWidget


class JsonEditor(Qsci.QsciScintilla):
    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.setLexer(Qsci.QsciLexerJSON(self))



from typing import Type, Union

from qgis.core import QgsApplication, QgsFields
from qgis.gui import QgsDateTimeEdit, QgsDoubleSpinBox, QgsSpinBox
from qgis.PyQt.QtCore import QVariant
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QCheckBox, QComboBox, QDateEdit, QWidget

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"


# noinspection PyCallByClass,PyArgumentList
def variant_type_icon(field_type: QVariant) -> QIcon:
    if field_type == QVariant.Bool:
        return QgsApplication.getThemeIcon("/mIconFieldBool.svg")
    elif field_type in [
        QVariant.Int,
        QVariant.UInt,
        QVariant.LongLong,
        QVariant.ULongLong,
    ]:
        return QgsApplication.getThemeIcon("/mIconFieldInteger.svg")
    elif field_type == QVariant.Double:
        return QgsApplication.getThemeIcon("/mIconFieldFloat.svg")
    elif field_type == QVariant.String:
        return QgsApplication.getThemeIcon("/mIconFieldText.svg")
    elif field_type == QVariant.Date:
        return QgsApplication.getThemeIcon("/mIconFieldDate.svg")
    elif field_type == QVariant.DateTime:
        return QgsApplication.getThemeIcon("/mIconFieldDateTime.svg")
    elif field_type == QVariant.Time:
        return QgsApplication.getThemeIcon("/mIconFieldTime.svg")
    elif field_type == QVariant.ByteArray:
        return QgsApplication.getThemeIcon("/mIconFieldBinary.svg")
    else:
        return QIcon()


def widget_for_field(field_type: QVariant) -> QWidget:
    q_combo_box = QComboBox()
    q_combo_box.setEditable(True)

    if field_type == QVariant.Bool:
        return QCheckBox()
    elif field_type in [
        QVariant.Int,
        QVariant.UInt,
        QVariant.LongLong,
        QVariant.ULongLong,
    ]:
        spin_box = QgsSpinBox()
        spin_box.setMaximum(2147483647)
        return spin_box
    elif field_type == QVariant.Double:
        spin_box = QgsDoubleSpinBox()
        spin_box.setMaximum(2147483647)
        return spin_box
    elif field_type == QVariant.String:
        return q_combo_box
    elif field_type == QVariant.Date:
        return QDateEdit()
    elif field_type == QVariant.DateTime:
        return QgsDateTimeEdit()
    elif field_type == QVariant.Time:
        return QgsDateTimeEdit()
    elif field_type == QVariant.ByteArray:
        return q_combo_box
    else:
        return q_combo_box


def value_for_widget(widget: Type[QWidget]) -> Union[str, bool, float, int]:
    if isinstance(widget, QComboBox):
        return widget.currentText()
    elif isinstance(widget, QCheckBox):
        return widget.isChecked()
    elif isinstance(widget, QgsDateTimeEdit):
        return widget.dateTime().toString("yyyy-MM-dd hh:mm:ss")
    elif isinstance(widget, QgsSpinBox) or isinstance(widget, QgsDoubleSpinBox):
        return widget.value()
    else:
        return str(widget.text())


def provider_fields(fields: QgsFields) -> QgsFields:
    flds = QgsFields()
    for i in range(fields.count()):
        if fields.fieldOrigin(i) == QgsFields.OriginProvider:
            flds.append(fields.at(i))
    return flds



# coding=utf-8
"""Common functionality used by regression tests."""

import os
import time
import warnings
from typing import Type, Union

from qgis.core import QgsApplication, QgsTask
from qgis.PyQt.QtCore import QCoreApplication

from ..tools.exceptions import QgsPluginNotImplementedException
from ..tools.tasks import BaseTask


def get_qgis_app():  # noqa
    warnings.warn(
        "get_qgis_app() is deprecated. Use library pytest-qgis instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    raise QgsPluginNotImplementedException(
        "get_qgis_app() is deprecated. Use library pytest-qgis instead."
    )


def is_running_inside_ci() -> bool:
    """Tells whether the plugin is running in CI environment"""
    return int(os.environ.get("QGIS_PLUGIN_IN_CI", "0")) == 1


def is_running_in_tools_module_ci() -> bool:
    return (
        is_running_inside_ci()
        and int(os.environ.get("QGIS_PLUGIN_TOOLS_IN_CI", "0")) == 1
    )


def qgis_supports_temporal() -> bool:
    try:
        from qgis.core import QgsRasterLayerTemporalProperties  # noqa F401

        return True
    except ImportError:
        return False


class TestTaskRunner:
    """
    This utility class could be used when running tasks in tests.
    """

    success = False
    fail = False
    progress = 0.0

    def completed(self) -> None:
        self.success = True

    def terminated(self) -> None:
        self.fail = True

    def set_progress(self, progress: float) -> None:
        self.progress = progress

    def run_task(
        self,
        task: QgsTask,
        cancel: bool = False,
        sleep_before_cancel: Union[int, float] = 0.0,
    ) -> bool:
        """
        Run task and return whether it was successful or not.
        """
        task.taskCompleted.connect(self.completed)
        task.taskTerminated.connect(self.terminated)
        task.progressChanged.connect(self.set_progress)
        QgsApplication.taskManager().addTask(task)

        if cancel:
            time.sleep(sleep_before_cancel)
            task.cancel()

        while not self.success and not self.fail:
            QCoreApplication.processEvents()

        return self.success


class SimpleTask(BaseTask):
    """
    Test task to used in tests needing a simple task.
    """

    def __init__(
        self,
        will_fail: bool = False,
        error_to_raise: Type[Exception] = ValueError,
        steps: int = 10,
        sleep_time: float = 0.01,
    ) -> None:
        super().__init__()
        self._will_fail = will_fail
        self._error_to_raise = error_to_raise
        self._steps = steps
        self._sleep_time = sleep_time

    def _run(self) -> bool:
        for i in range(self._steps):
            self.setProgress(i * self._steps)
            if self._will_fail:
                raise self._error_to_raise("custom failure")
            self._check_if_canceled()
            time.sleep(self._sleep_time)
        return True



import logging
import sys
from typing import Any, Dict, Optional

from .custom_logging import bar_msg


class MessageBarLogger:
    """
    logging.Logger like interface to push messages to the
    message bar where necessary with info/warning/etc methods.

    Setup with a logger name that has a message bar set.
    """

    def __init__(self, logger_name: str) -> None:
        self._logger = logging.getLogger(logger_name)
        self._logger_kwargs: Dict[str, Any] = (
            {}
            if sys.version_info.major == 3 and sys.version_info.minor < 8
            else {"stacklevel": 2}
        )

    def info(
        self,
        message: Any,
        details: Any = "",
        duration: Optional[int] = None,
        success: bool = False,
        exc_info: Optional[Exception] = None,
        stack_info: bool = False,
    ) -> None:
        """
        Logs info messages to message bar and to other logging handlers
        :param message: Header of the message
        :param details: Longer body of the message. Can be set to empty string.
        :param duration: can be used to specify the message timeout in seconds. If
            ``duration`` is set to 0, then the message must be manually dismissed
            by the user.
        :param success: Whether the message is success message or not
        :param exc_info: Exception of handled exception for capturing traceback
        :param stack_info: Whether to include stack info
        """

        self._logger.info(
            str(message),
            extra=bar_msg(details, duration, success),
            exc_info=exc_info,
            stack_info=stack_info,
            **self._logger_kwargs,
        )
        if details != "":
            self._logger.info(
                str(details),
                exc_info=exc_info,
                stack_info=stack_info,
                **self._logger_kwargs,
            )

    def warning(
        self,
        message: Any,
        details: Any = "",
        duration: Optional[int] = None,
        success: bool = False,
        exc_info: Optional[Exception] = None,
        stack_info: bool = False,
    ) -> None:
        """
        Logs warning messages to message bar and to other logging handlers
        :param message: Header of the message
        :param details: Longer body of the message. Can be set to empty string.
        :param duration: can be used to specify the message timeout in seconds. If
            ``duration`` is set to 0, then the message must be manually dismissed
            by the user.
        :param success: Whether the message is success message or not
        :param exc_info: Exception of handled exception for capturing traceback
        :param stack_info: Whether to include stack info
        """
        self._logger.warning(
            str(message),
            extra=bar_msg(details, duration, success),
            exc_info=exc_info,
            stack_info=stack_info,
            **self._logger_kwargs,
        )
        if details != "":
            self._logger.warning(
                str(details),
                exc_info=exc_info,
                stack_info=stack_info,
                **self._logger_kwargs,
            )

    def error(
        self,
        message: Any,
        details: Any = "",
        duration: Optional[int] = None,
        success: bool = False,
        exc_info: Optional[Exception] = None,
        stack_info: bool = False,
    ) -> None:
        """
        Logs error of risen exception to message bar and to other logging handlers
        :param message: Header of the message
        :param details: Longer body of the message. Can be set to empty string.
        :param duration: can be used to specify the message timeout in seconds. If
            ``duration`` is set to 0, then the message must be manually dismissed
            by the user.
        :param success: Whether the message is success message or not
        :param exc_info: Exception of handled exception for capturing traceback
        :param stack_info: Whether to include stack info
        """
        self._logger.error(
            str(message),
            extra=bar_msg(details, duration, success),
            exc_info=exc_info,
            stack_info=stack_info,
            **self._logger_kwargs,
        )
        if details != "":
            self._logger.error(
                str(details),
                exc_info=exc_info,
                stack_info=stack_info,
                **self._logger_kwargs,
            )

    def exception(
        self,
        message: Any,
        details: Any = "",
        duration: Optional[int] = None,
        success: bool = False,
        exc_info: Optional[Exception] = None,
        stack_info: bool = False,
    ) -> None:
        """
        Logs error with traceback of risen exception to message bar and to
        other logging handlers
        :param message: Header of the message
        :param details: Longer body of the message. Can be set to empty string.
        :param duration: can be used to specify the message timeout in seconds. If
            ``duration`` is set to 0, then the message must be manually dismissed
            by the user.
        :param success: Whether the message is success message or not
        :param exc_info: Exception of handled exception for capturing traceback
        :param stack_info: Whether to include stack info
        """
        self._logger.exception(
            str(message),
            extra=bar_msg(details, duration, success),
            stack_info=stack_info,
            **self._logger_kwargs,
        )
        if details != "":
            self._logger.error(
                str(details),
                exc_info=exc_info,
                stack_info=stack_info,
                **self._logger_kwargs,
            )


# publish the old MsgBar with the default logger name
MsgBar = MessageBarLogger(__name__)



"""Tools to work with resource files."""
import configparser
import importlib.resources
import inspect
import sys
from os.path import abspath, dirname, exists, join, pardir
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Iterator, NamedTuple, Optional

from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QDialog, QWidget

if TYPE_CHECKING:
    import os
    from typing import Union

__copyright__ = (
    "Copyright 2019, 3Liz, 2020-2021 Gispo Ltd, "
    "2022-2023 National Land Survey of Finland"
)
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"

from qgis.core import QgsApplication

PLUGIN_NAME: str = ""
SLUG_NAME: str = ""


_TOP_LEVEL_NAME = __name__.split(".", maxsplit=1)[0]
_IS_SUBMODULE_USAGE = not _TOP_LEVEL_NAME == "qgis_plugin_tools"


def _plugin_path_submodule() -> str:
    # assume qgis_plugin_tools is inside the plugin package,
    # use the path to the top level module name

    module_file = sys.modules[_TOP_LEVEL_NAME].__file__

    if module_file is not None:
        path = str(Path(module_file).parent.resolve())
    else:
        # maybe possible to have __file__ as none? fall back to default
        # structure with qgis_plugin_tools directly under plugin package
        path = dirname(dirname(__file__))
        path = abspath(abspath(join(path, pardir)))

    return path


def _iterate_modules(module_name: str) -> Iterator[str]:
    """Iterates modules bottom up

    >>> list(_iterate_modules("plugin.tools.resources"))
    [
        "plugin.tools.resources",
        "plugin.tools",
        "plugin"
    ]
    """

    modules = module_name.split(".")
    for i in range(len(modules), 0, -1):
        yield ".".join(modules[:i])


class IsPluginResult(NamedTuple):
    is_plugin: bool
    plugin_directory: Optional[str] = None

    def __bool__(self) -> bool:
        return self.is_plugin


def _is_module_qgis_plugin(module_name: str) -> IsPluginResult:
    """Checks if the module is a QGIS plugin

    A plugin module has a classFactory method and a metadata.txt at the package root.

    Returns a IsPluginResult thats compares to True when the module given is a plugin.

    >>> if is_plugin := _is_module_qgis_plugin('myplugin'):
    >>>    print(f"Plugin is installed ad {is_plugin.plugin_directory}")
    """

    module = sys.modules.get(module_name)
    if module is None or not inspect.ismodule(module):
        return IsPluginResult(False)

    class_factory_function = getattr(module, "classFactory", None)
    if class_factory_function is None or not inspect.isfunction(class_factory_function):
        return IsPluginResult(False)

    try:
        source_file = inspect.getsourcefile(module)
    except TypeError:
        # TypeError is thrown for built-in modules. We are only interested in
        # custom modules, so it is safe to ignore the error.
        return IsPluginResult(False)
    if not source_file:
        return IsPluginResult(False)

    source_directory = dirname(source_file)
    if not exists(join(source_directory, "metadata.txt")):
        return IsPluginResult(False)

    return IsPluginResult(True, source_directory)


def _plugin_path_dependency() -> str:
    """Get the path to the plugin package folder.

    Goes up the call stack and up the parent packages until package is
    recognized as QGIS plugin.
    """

    for frame_info in inspect.stack():
        caller_module_name: Optional[str] = frame_info.frame.f_globals.get("__name__")
        if caller_module_name is None:
            continue
        for module_name in _iterate_modules(caller_module_name):
            if is_plugin := _is_module_qgis_plugin(module_name):
                assert is_plugin.plugin_directory
                return is_plugin.plugin_directory

    # fall back to default directory tree
    return _plugin_path_submodule()


def plugin_path(*args: str) -> str:
    """Get the path to plugin package folder.

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    :type args: str

    :return: Absolute path to the resource.
    :rtype: str
    """
    if _IS_SUBMODULE_USAGE:
        path = _plugin_path_submodule()
    else:
        path = _plugin_path_dependency()

    for item in args:
        path = abspath(join(path, item))

    return path


def root_path(*args: str) -> str:
    """Get the path to plugin root folder.

    NOTE: the assumed root is the parent of the plugin package folder.

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    :type args: str

    :return: Absolute path to the resource.
    :rtype: str
    """
    return plugin_path(pardir, *args)


def profile_path(*args: str) -> str:
    """
    Get the path inside profile folder.
    return: Absolute path to the resource.
    """
    path = QgsApplication.qgisSettingsDirPath()
    for item in args:
        path = abspath(join(path, item))

    return path


def plugin_name() -> str:
    """Return the plugin name according to metadata.txt but stripped of spaces.
    This is because lots of the code assumes plugin_name returns a string without
    spaces.

    :return: The stripped plugin name.
    :rtype: basestring
    """
    global PLUGIN_NAME, _IS_SUBMODULE_USAGE

    if PLUGIN_NAME != "":
        return PLUGIN_NAME

    try:
        metadata = metadata_config()
        name: str = metadata["general"]["name"]
        name = name.replace(" ", "").strip()
    except KeyError:
        name = "test_plugin"

    # if qgis plugin tools is run as a dependency, global var cannot be set
    # since it might confuse multiple plugins in the same env using this fn
    if _IS_SUBMODULE_USAGE:
        PLUGIN_NAME = name

    return name


def plugin_display_name() -> str:
    """Return the plugin name according to metadata.txt, including spaces,
    for displaying to the end user.

    :return: The original plugin name.
    :rtype: basestring
    """
    try:
        metadata = metadata_config()
        return metadata["general"]["name"]
    except KeyError:
        return "Test plugin"


def slug_name() -> str:
    """Return project slug name in .qgis-plugin.ci"""
    global SLUG_NAME, _IS_SUBMODULE_USAGE

    if SLUG_NAME != "":
        return SLUG_NAME

    try:
        metadata = metadata_config()
        name: str = metadata["general"]["repository"]
        slug = name.split("/")[-1]
    except KeyError:
        slug = plugin_name()

    # if qgis plugin tools is run as a dependency, global var cannot be set
    # since it might confuse multiple plugins in the same env using this fn
    if _IS_SUBMODULE_USAGE:
        SLUG_NAME = slug

    return slug


def task_logger_name() -> str:
    """
    Returns the name for task logger
    """
    return f"{plugin_name()}_task"


def metadata_config() -> configparser.ConfigParser:
    """Get the INI config parser for the metadata file.

    :return: The config parser object.
    :rtype: ConfigParser
    """
    path = plugin_path("metadata.txt")
    config = configparser.ConfigParser()
    config.read(path)
    return config


def qgis_plugin_ci_config() -> Optional[Dict]:
    """
    Get configuration of the ci config or None
    """
    path_str = root_path(".qgis-plugin-ci")
    if not Path(path_str).exists():
        path_str = plugin_path(".qgis-plugin-ci")
    path = Path(path_str)
    if path.exists():
        with open(path) as f:
            config = {}
            for line in f:
                parts = line.split(":")
                config[parts[0]] = ":".join(parts[1:])

        return config
    return None


def plugin_test_data_path(*args: str) -> str:
    """Get the path to the plugin test data path.

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    :type args: str

    :return: Absolute path to the resources folder.
    :rtype: str
    """

    path = abspath(abspath(join(root_path(), "test", "data")))
    if not exists(path):
        path = abspath(abspath(join(plugin_path(), "test", "data")))
    for item in args:
        path = abspath(join(path, item))

    return path


def resources_path(*args: str) -> str:
    """Get the path to our resources folder.

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    :type args: str

    :return: Absolute path to the resources folder.
    :rtype: str
    """
    path = abspath(abspath(join(plugin_path(), "resources")))
    for item in args:
        path = abspath(join(path, item))

    return path


def qgis_plugin_tools_resources(*args: str) -> str:
    """Get the path within the qgis_plugin_tools submodule"""
    return str(Path(__file__, "..", "..", "resources", *args).resolve().absolute())


def load_ui(*args: str) -> QWidget:
    """Get compiled UI file.

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    :type args: str

    :return: Compiled UI file.
    """
    ui_file_path = resources_path("ui", *args)
    return load_ui_from_file(ui_file_path)


def load_ui_from_file(ui_file_path: "Union[str, os.PathLike]") -> QWidget:
    ui_class, _ = uic.loadUiType(ui_file_path)
    return ui_class


def ui_file_dialog(*ui_file_name_parts: str):  # noqa ANN201
    """DRY helper for building classes from a .ui file"""

    class UiFileDialogClass(QDialog, load_ui(*ui_file_name_parts)):  # type: ignore
        def __init__(
            self,
            parent: Optional[QWidget],
        ) -> None:
            super().__init__(parent)
            self.setupUi(self)  # provided by load_ui FORM_CLASS

    return UiFileDialogClass


def package_file(package: importlib.resources.Package, file_name: str) -> Path:
    """
    Safely access a file in the package hierarchy. This will
    ensure the requested file actually exists on the file system
    outside the contextmanager, so that the file can be still
    accessed with the path on demand later.

    Use like importlib.resources, provide a package (module or string)
    and file name/path inside the package.

    >>> package_file('myplugin.resources', 'image.svg')
    C:/Users/me/AppData/Roaming/QGIS/QGIS3/profiles/default/python/plugins/myplugin/resources/image.svg

    >>> import myplugin.resources
    >>> package_file(myplugin.resources, 'image.svg')
    C:/Users/me/AppData/Roaming/QGIS/QGIS3/profiles/default/python/plugins/myplugin/resources/image.svg
    """

    with importlib.resources.path(package, file_name) as requested_path:
        if not requested_path.is_file():
            raise FileNotFoundError(
                f"requested file {file_name} not found in {package}"
            )

    if not requested_path.is_file():
        raise FileNotFoundError(
            "requested file would be available only as a temporary resource"
        )

    return requested_path



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 2"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

from typing import Any, Dict, Optional

from qgis.PyQt.QtNetwork import QNetworkReply

from .i18n import tr


class QgsPluginException(Exception):
    """Use this as a base exception class in custom exceptions"""

    # Override default_msg to set default message in inherited classes
    default_msg = ""

    def __init__(
        self, message: Optional[str] = None, bar_msg: Optional[Dict[str, Any]] = None
    ) -> None:
        """
        Initializes the exception with custom bar_msg to be shown in message bar
        :param message: Title of the message
        :param bar_msg: dictionary formed by tools.custom_logging.bar_msg
        """
        if message is None:
            message = self.default_msg
        self.message = message
        super().__init__(message)
        self.bar_msg: Dict[str, Any] = bar_msg if bar_msg is not None else {}


class QgsPluginNetworkException(QgsPluginException):
    default_msg = tr("A network error occurred.")

    def __init__(
        self,
        *args: Any,
        error: Optional[QNetworkReply.NetworkError] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the exception with error details so the plugin may process
        different network exceptions differently.
        :param status: The QNetworkReply error type
        """
        self.error = error
        super().__init__(*args, **kwargs)


class QgsPluginNotImplementedException(QgsPluginException):
    pass


class QgsPluginVersionInInvalidFormat(QgsPluginException):
    pass


class QgsPluginInvalidProjectSetting(QgsPluginException):
    pass


class QgsPluginExpressionException(QgsPluginException):
    default_msg = tr("There is an error in the expression")


class TaskInterruptedException(QgsPluginException):
    pass



from ..tools.i18n import tr


def test_tr_formatting():
    string = tr(
        "These are args {} {} and these are kwargs {foo} {bar}", 1, 2, foo=3, bar=4
    )
    assert string == "These are args 1 2 and these are kwargs 3 4"



# flake8: noqa N802
import logging

from qgis.core import QgsProcessingFeedback

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

LOGGER = logging.getLogger(__name__)


class LoggerProcessingFeedBack(QgsProcessingFeedback):
    def __init__(self, use_logger=False):
        super().__init__()
        self._last = None
        self.use_logger = use_logger
        self.last_progress_text = None
        self.last_push_info = None
        self.last_command_info = None
        self.last_debug_info = None
        self.last_console_info = None
        self.last_report_error = None

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, text):
        self._last = text

    def setProgressText(self, text):
        self._last = text
        self.last_progress_text = text
        if self.use_logger:
            LOGGER.info(text)

    def pushInfo(self, text):
        self._last = text
        self.last_push_info = text
        if self.use_logger:
            LOGGER.info(text)

    def pushCommandInfo(self, text):
        self._last = text
        self.last_command_info = text
        if self.use_logger:
            LOGGER.info(text)

    def pushDebugInfo(self, text):
        self._last = text
        self.last_debug_info = text
        if self.use_logger:
            LOGGER.warning(text)

    def pushConsoleInfo(self, text):
        self._last = text
        self.last_console_info = text
        if self.use_logger:
            LOGGER.info(text)

    def reportError(self, text, fatalError=False):
        self._last = text
        self.last_report_error = text
        if self.use_logger:
            LOGGER.exception(text)



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

from qgis.core import QgsRectangle


def extent_to_bbox(extent: QgsRectangle, precision: int = 2) -> str:
    """
    Add extent for the query

    :param extent: QgsRectangle expected to be in the right extent
    :param precision: Precision of coordinates
    :return: string representation xmin,ymin,xmax,ymax
    """

    def rnd(c: float) -> float:
        return round(c, precision)

    bbox = (
        rnd(extent.xMinimum()),
        rnd(extent.yMinimum()),
        rnd(extent.xMaximum()),
        rnd(extent.yMaximum()),
    )
    return ",".join(map(str, bbox))



"""Tests QGIS plugin init. Modified from unittest to pytest and using plugin_path"""

__author__ = "Tim Sutton <tim@linfiniti.com>"
__revision__ = "$Format:%H$"
__date__ = "17/10/2010"
__license__ = "GPL"
__copyright__ = "Copyright 2012, Australia Indonesia Facility for "

import configparser

import pytest

from ..testing.utilities import is_running_in_tools_module_ci
from ..tools.resources import plugin_path

__copyright__ += "Disaster Reduction"


@pytest.mark.skipif(is_running_in_tools_module_ci(), reason="In CI")
def test_read_init():
    """Test that the plugin __init__ will validate on plugins.qgis.org."""

    # You should update this list according to the latest in
    # https://github.com/qgis/qgis-django/blob/master/qgis-app/
    #        plugins/validator.py

    required_metadata = [
        "name",
        "description",
        "version",
        "qgisMinimumVersion",
        "email",
        "author",
    ]

    file_path = plugin_path("metadata.txt")
    metadata = []
    parser = configparser.ConfigParser()
    parser.optionxform = str
    parser.read(file_path)
    message = 'Cannot find a section named "general" in %s' % file_path
    assert parser.has_section("general"), message
    metadata.extend(parser.items("general"))

    for expectation in required_metadata:
        message = 'Cannot find metadata "%s" in metadata source (%s).' % (
            expectation,
            file_path,
        )

        assert expectation in dict(metadata), message



__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import datetime

from qgis.core import (
    QgsContrastEnhancement,
    QgsDateTimeRange,
    QgsRasterBandStats,
    QgsRasterDataProvider,
    QgsRasterLayer,
    QgsSingleBandGrayRenderer,
)
from qgis.PyQt.QtCore import Qt

try:
    from qgis.core import QgsRasterLayerTemporalProperties
except ImportError:
    QgsRasterLayerTemporalProperties = None


def set_raster_renderer_to_singleband(layer: QgsRasterLayer, band: int = 1) -> None:
    """
    Set raster renderer to singleband
    :param layer: raster layer
    :param band: band number starting from 1
    """
    # https://gis.stackexchange.com/a/377631/123927 and
    # https://gis.stackexchange.com/a/157573/123927
    provider: QgsRasterDataProvider = layer.dataProvider()
    renderer: QgsSingleBandGrayRenderer = QgsSingleBandGrayRenderer(
        layer.dataProvider(), band
    )

    stats: QgsRasterBandStats = provider.bandStatistics(
        band, QgsRasterBandStats.All, layer.extent(), 0
    )
    min_val = max(stats.minimumValue, 0)
    max_val = max(stats.maximumValue, 0)

    enhancement = QgsContrastEnhancement(renderer.dataType(band))
    contrast_enhancement = QgsContrastEnhancement.StretchToMinimumMaximum
    enhancement.setContrastEnhancementAlgorithm(contrast_enhancement, True)
    enhancement.setMinimumValue(min_val)
    enhancement.setMaximumValue(max_val)
    layer.setRenderer(renderer)
    layer.renderer().setContrastEnhancement(enhancement)
    layer.triggerRepaint()


def set_fixed_temporal_range(layer: QgsRasterLayer, t_range: QgsDateTimeRange) -> None:
    """
    Set fixed temporal range for raster layer
    :param layer: raster layer
    :param t_range: fixed temporal range
    """
    mode = QgsRasterLayerTemporalProperties.ModeFixedTemporalRange
    tprops: QgsRasterLayerTemporalProperties = layer.temporalProperties()
    tprops.setMode(mode)
    if t_range.begin().timeSpec() == 0 or t_range.end().timeSpec() == 0:
        begin = t_range.begin()
        end = t_range.end()
        begin.setTimeSpec(Qt.TimeSpec(1))
        end.setTimeSpec(Qt.TimeSpec(1))
        t_range = QgsDateTimeRange(begin, end)
    tprops.setFixedTemporalRange(t_range)
    tprops.setIsActive(True)


def set_band_based_on_range(layer: QgsRasterLayer, t_range: QgsDateTimeRange) -> int:
    """
    Set raster layer band based on temporal range
    :param layer: Raster layer
    :param t_range: temporal range
    :return: band number
    """
    band_num = 1
    tprops: QgsRasterLayerTemporalProperties = layer.temporalProperties()
    if (
        tprops.isVisibleInTemporalRange(t_range)
        and t_range.begin().isValid()
        and t_range.end().isValid()
    ):
        if tprops.mode() == QgsRasterLayerTemporalProperties.ModeFixedTemporalRange:
            layer_t_range: QgsDateTimeRange = tprops.fixedTemporalRange()
            start: datetime.datetime = layer_t_range.begin().toPyDateTime()
            end: datetime.datetime = layer_t_range.end().toPyDateTime()
            delta = (end - start) / layer.bandCount()
            band_num = int((t_range.begin().toPyDateTime() - start) / delta) + 1
            set_raster_renderer_to_singleband(layer, band_num)
    return band_num



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa
import argparse
import os
import shutil
import subprocess
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import List
from zipfile import ZipFile

from ..tools.resources import plugin_name, plugin_path, resources_path

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"


def is_windows():
    return "win" in sys.platform and "darwin" not in sys.platform


PLUGINNAME = plugin_name()

PLUGIN_PACKAGE_NAME = Path(__file__).parent.parent.parent.resolve().name
ROOT_DIR = str(Path(__file__).parent.parent.parent.parent.resolve())

SUBMODULES = ["qgis_plugin_tools"]

# Add files for any locales you want to support here
LOCALES: List[str] = []

# If locales are enabled, set the name of the lrelease binary on your system. If
# you have trouble compiling the translations, you may have to specify the full path to
# lrelease
LRELEASE = "lrelease"  # 'lrelease-qt4'

PYRCC = "pyrcc5"

# Name of the QGIS profile you are using in development
PROFILE = "default"

# Resource files
RESOURCES_SRC: List[str] = []

EXTRAS = ["metadata.txt"]

EXTRA_DIRS = ["resources"]

COMPILED_RESOURCE_FILES = ["resources.py"]

VENV_NAME = ".venv"

"""
#################################################
# Normally you would not need to edit below here
#################################################
"""

STARTUP_BAT = r"""
@echo off
set IDE={ide}
set REPOSITORY={repository}
set OSGEO4W_ROOT={qgis_root}
set QGIS_PREFIX_PATH={qgis_prefix_path}
set TEMP_PATH=%PATH%
call "%OSGEO4W_ROOT%"\bin\o4w_env.bat
call "%OSGEO4W_ROOT%"\bin\qt5_env.bat
call "%OSGEO4W_ROOT%"\bin\py3_env.bat
call "%OSGEO4W_ROOT%"\apps\grass\grass78\etc\env.bat
path %QGIS_PREFIX_PATH%\bin;%PATH%
path %PATH%;%OSGEO4W_ROOT%\apps\Qt5\bin
path %PATH%;%OSGEO4W_ROOT%\apps\Python37\Scripts
path %QGIS_PREFIX_PATH%\bin;%PATH%
:: Add original PATH
path %PATH%;%TEMP_PATH%
set GDAL_FILENAME_IS_UTF8=YES
:: Set VSI cache to be used as buffer, see #6448
set VSI_CACHE=TRUE
set VSI_CACHE_SIZE=1000000
set QT_PLUGIN_PATH=%QGIS_PREFIX_PATH%\qtplugins;%OSGEO4W_ROOT%\apps\Qt5\plugins
set PYTHONPATH=%QGIS_PREFIX_PATH%\python;%OSGEO4W_ROOT%\apps\Qt5\plugins;%PYTHONPATH%

start "Start your IDE aware of QGIS" /B %IDE% %REPOSITORY%
"""

VENV_CREATION_SCRIPT = """
python -m venv --system-site-packages --clear {venv_path}
{qgis_path_fix}
{source}{activator}
python -m pip install --upgrade pip
python -m pip install -r {requirements}
pre-commit install
"""

# self.qgis_dir points to the location where your plugin should be installed.
# This varies by platform, relative to your HOME directory:
# 	* Linux:
# 	  .local/share/QGIS/QGIS3/profiles/default/python/plugins/
# 	* Mac OS X:
# 	  Library/Application Support/QGIS/QGIS3/profiles/default/python/plugins
# 	* Windows:
# 	  AppData\Roaming\QGIS\QGIS3\profiles\default\python\plugins'

if sys.platform == "linux":
    dr = os.path.join(".local", "share")
elif is_windows():
    dr = os.path.join("AppData", "Roaming")
else:
    dr = os.path.join("Library", "Application Support")

VERBOSE = False


def echo(*args, **kwargs):
    if VERBOSE or kwargs.get("force", False):
        print(*args)


class PluginMaker:
    def __init__(
        self,
        py_files,
        ui_files,
        resources=RESOURCES_SRC,
        extra_dirs=EXTRA_DIRS,
        extras=EXTRAS,
        compiled_resources=COMPILED_RESOURCE_FILES,
        locales=LOCALES,
        profile=PROFILE,
        lrelease=LRELEASE,
        pyrcc=PYRCC,
        verbose=VERBOSE,
        submodules=SUBMODULES,
    ):
        global VERBOSE
        self.py_files = py_files
        self.ui_files = ui_files
        self.resources = resources
        self.extra_dirs = extra_dirs
        self.extras = extras
        self.compiled_resources = compiled_resources
        self.locales = locales
        self.profile = profile
        self.lrelease = lrelease
        self.pyrcc = pyrcc
        self.qgis_dir = os.path.join(dr, "QGIS", "QGIS3", "profiles", profile)
        self.plugin_dir = os.path.join(
            str(Path.home()), self.qgis_dir, "python", "plugins", PLUGIN_PACKAGE_NAME
        )
        self.submodules = submodules
        VERBOSE = verbose

        # git-like usage https://chase-seibert.github.io/blog/2014/03/21/python-multilevel-argparse.html # noqa
        usage = f"""build.py <command> [<args>]
Commands:
     clean          Cleans resources
     compile        Compiles resources to resources.py
     deploy         Deploys the plugin to the QGIS plugin directory ({self.plugin_dir})
     package        Builds a package that can be uploaded to Github releases
                    or to the plugin repository
     start_ide      Start IDE of your choice. This is required on Windows to
                    assure that the environment is set correctly.
     transup        Search for new strings to be translated
     transcompile   Compile translation files to .qm files.
     venv           Create python virtual environment
Put -h after command to see available optional arguments if any
"""
        parser = ArgumentParser(usage=usage)
        parser.add_argument("command", help="Subcommand to run")
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            parser.print_help()
            exit(1)

        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def clean(self):
        for fil in self.compiled_resources:
            if os.path.exists(fil):
                echo(f"rm {fil}")
                os.remove(fil)

    def compile(self):
        pre_args = self._get_platform_args()
        for fil in self.resources:
            if os.path.exists(fil):
                args = pre_args + [self.pyrcc, "-o", fil.replace(".qrc", ".py"), fil]
                self.run_command(args)
            else:
                raise ValueError(f"The expected resource file {fil} is missing!")

    def _get_platform_args(self):
        pre_args = []
        if is_windows():
            pre_args = ["cmd", "/c"]  # noqa W605
        return pre_args

    def deploy(self):
        self.compile()
        dst_dir = f"{self.plugin_dir}/"
        os.makedirs(self.plugin_dir, exist_ok=True)
        for dr in self.extra_dirs:
            echo(f"cp -R --parents {dr} {dst_dir}")
            dst = os.path.join(self.plugin_dir, dr)
            if os.path.exists(dst):
                shutil.rmtree(dst)
            shutil.copytree(dr, dst)
        self.cp_parents(dst_dir, self.extras)
        self.cp_parents(dst_dir, self.compiled_resources)
        self.cp_parents(dst_dir, self.py_files)
        self.cp_parents(dst_dir, self.ui_files)

    def package(self):
        parser = ArgumentParser()
        parser.add_argument(
            "--version",
            type=str,
            help="Version number of the tag (eg. --version v0.0.1",
        )
        parser.add_argument(
            "--tag",
            action="store_true",
            help="Run git tag as well. REMEMBER to update metadata.txt with "
            "your version before this",
        )
        parser.set_defaults(test=False)
        args = parser.parse_args(sys.argv[2:])
        if args.version is None:
            echo("Give valid version number", force=True)
            parser.print_help()
            exit(1)

        if args.tag:
            self.run_command(self._get_platform_args() + ["git", "tag", args.version])

        pkg_command = [
            "git",
            "archive",
            f"--prefix={PLUGINNAME}/",
            "-o",
            f"{PLUGINNAME}.zip",
            args.version,
        ]
        self.run_command(self._get_platform_args() + pkg_command)

        for submodule in self.submodules:
            d = plugin_path(submodule)
            pkg_command = [
                "git",
                "archive",
                f"--prefix={PLUGINNAME}/{submodule}/",
                "-o",
                f"{submodule}.zip",
                "master",
            ]
            self.run_command(self._get_platform_args() + pkg_command, d=d)
        zips = [f"{PLUGINNAME}.zip"] + [
            os.path.abspath(os.path.join(plugin_path(submodule), f"{submodule}.zip"))
            for submodule in self.submodules
        ]
        self.join_zips(zips)
        echo(f"Created package: {PLUGINNAME}.zip")

    def start_ide(self):
        if not is_windows():
            print(
                "This command is only meant to run on Windows environment with QGIS < 3.16.8."
            )
            return
        parser = ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument(
            "--ide",
            type=str,
            help=r"Path to your .exe (eg. C:\Program Files\JetBrains\IntelliJ IDEA 2020.3.2\bin\idea64.exe)."
            "Set the path to a environment variable QGIS_DEV_IDE to avoid typing it each time.",
            default=os.environ.get("QGIS_DEV_IDE", ""),
        )
        parser.add_argument(
            "--qgis_root",
            type=str,
            help=r"Path to your QGIS installation directory. (eg. C:\OSGeo4W64 or C:\QGIS_3_16). "
            r"Set the path to a environment variable QGIS_DEV_OSGEO4W_ROOT to avoid typing it each time.",
            default=os.environ.get("QGIS_DEV_OSGEO4W_ROOT", r"C:\OSGeo4W64"),
        )
        parser.add_argument(
            "--qgis_prefix_path",
            type=str,
            help=r"This is a path to OSGEO4W_ROOT/apps/ which is either qgis or qgis-ltr. "
            r"Set the path to a environment variable QGIS_DEV_PREFIX_PATH to avoid typing it each time.",
            default=os.environ.get(
                "QGIS_DEV_PREFIX_PATH", r"C:\OSGeo4W64\apps\qgis-ltr"
            ),
        )
        parser.add_argument(
            "--save_to_disk",
            action="store_true",
            help=r"Saves the startup script  to the disk rather than starting the IDE.",
        )
        args = parser.parse_args(sys.argv[2:])
        if not args.ide:
            print("Add reasonable value to --ide")
            return

        script = STARTUP_BAT.format(
            ide=args.ide if '"' in args.ide or " " not in args.ide else f'"{args.ide}"',
            repository=ROOT_DIR,
            qgis_root=args.qgis_root,
            qgis_prefix_path=args.qgis_prefix_path,
        )

        if args.save_to_disk:
            with open(Path(ROOT_DIR) / "start_ide.bat", "w") as f:
                f.write(script)
                print(
                    f"Script saved successfully to {f.name}. "
                    f"You can move the file whenever you want."
                )
        else:
            process = subprocess.Popen(
                "cmd.exe",
                shell=False,
                universal_newlines=True,
                stdin=subprocess.PIPE,
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
            process.communicate(script)

    def transup(self):
        files_to_translate = self.py_files + self.ui_files
        for locale in self.locales:
            ts_file = os.path.join(resources_path("i18n"), f"{locale}.ts")
            args = (
                self._get_platform_args()
                + ["pylupdate5", "-noobsolete"]
                + files_to_translate
                + ["-ts", ts_file]
            )
            self.run_command(args, force_show_output=True)

    def transcompile(self):
        pre_args = self._get_platform_args()
        for locale in self.locales:
            fil = os.path.join(resources_path("i18n"), f"{locale}.ts")
            echo(f"Processing {fil}")
            args = pre_args + [self.lrelease, fil]
            self.run_command(args, force_show_output=True)

    def venv(self):
        try:
            from qgis.core import QgsVectorLayer
        except ImportError:
            print("Your python environment has no access to QGIS libraries!")
            return

        if is_windows():
            env = os.environ.copy()
            env["PATH"] += (
                f';{os.path.join(os.path.expanduser("~"), "AppData", "Local", "Programs", "Git", "cmd")}'
                ";C:\\Program Files\\Git\\cmd"
            )
        else:
            env = os.environ

        print("Installing virtual environment")
        requirements = os.path.join(ROOT_DIR, "requirements-dev.txt")
        venv_path = os.path.join(ROOT_DIR, VENV_NAME)
        qgis_path_fix = (
            'python -c "import pathlib;'
            "import qgis;"
            "print(str((pathlib.Path(qgis.__file__)/'../..').resolve()))\" "
            f"> {os.path.join(venv_path, 'qgis.pth')}"
        )
        script = VENV_CREATION_SCRIPT.format(
            venv_path=venv_path,
            source="source " if not is_windows() else "",
            activator=os.path.join(
                ROOT_DIR,
                VENV_NAME,
                "bin" if not is_windows() else "Scripts",
                "activate",
            ),
            qgis_path_fix=qgis_path_fix if is_windows() else "",
            requirements=requirements,
        )

        process = subprocess.Popen(
            "cmd.exe" if is_windows() else "sh",
            shell=False,
            universal_newlines=True,
            stdin=subprocess.PIPE,
            stdout=sys.stdout,
            stderr=sys.stderr,
            env=env,
        )
        process.communicate(script)

    @staticmethod
    def run_command(args, d=None, force_show_output=False):
        cmd = " ".join(args)
        if d is not None:
            cmd = f"cd {d} && " + cmd
        echo(cmd, force=force_show_output)
        pros = subprocess.Popen(
            args,
            cwd=d,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        stdout, stderr = pros.communicate()
        echo(stdout, force=force_show_output)
        if len(stderr):
            echo(stderr, force=True)
            print(
                "------beging of stderr----------:\n",
                stderr,
                "\n-----end of stderr-----",
            )
            raise ValueError("Stopping now due to error in stderr!")

    @staticmethod
    def cp_parents(target_dir, files):
        """https://stackoverflow.com/a/15340518"""
        dirs = []
        for file in files:
            dirs.append(os.path.dirname(file))
        dirs.sort(reverse=True)
        for i in range(len(dirs)):
            if not dirs[i] + os.sep in dirs[i - 1]:
                need_dir = os.path.normpath(target_dir + dirs[i])
                echo("mkdir", need_dir)
                os.makedirs(need_dir, exist_ok=True)
        for file in files:
            dest = os.path.normpath(target_dir + file)
            echo(f"cp {file} {dest}")
            shutil.copy(file, dest)

    @staticmethod
    def join_zips(zips):
        """
        https://stackoverflow.com/a/10593823/10068922

        Open the first zip file as append and then read all
        subsequent zip files and append to the first one
        """
        with ZipFile(zips[0], "a") as z1:
            for fname in zips[1:]:
                zf = ZipFile(fname, "r")
                for n in zf.namelist():
                    z1.writestr(n, zf.open(n).read())



import importlib.resources
from typing import Any, Dict, Type

from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QWidget

from .resources import package_file


class CompiledUI:
    # provided by pyuic .ui file compilation
    def setupUi(self, instance: QWidget) -> None:  # noqa: N802
        ...


def load_ui_file(package: importlib.resources.Package, ui_file_name: str) -> Any:
    """
    Use like importlib.resources to load a single ui file from a package to a class:

    ```
    MyUi: Type[QDockWidget] = load_ui_file(my.imported.module, "dock_widget.ui")
    ```

    Type the return as a type of the ui file subclass, and build the class with
    type-ignore on the inheritance. Mypy cannot statically determine the dynamic
    base class without a plugin, but for example Pylance can still provide
    autocomplete for the implementation methods:

    ```
    class MyImplementation(MyUi):  # type: ignore
        def __init__(self):
            super().__init__()  # Pylance will show QDockWidget init signature
    ```
    """
    # TODO: add mypy plugin to support usage without type-ignores?

    ui_file_path = package_file(package, ui_file_name)

    ui_class: Type[CompiledUI]
    base_class: Type[QWidget]
    ui_class, base_class = uic.loadUiType(str(ui_file_path))

    class UiFileWidget(base_class, ui_class):  # type: ignore
        def __init__(
            self,
            *args: Any,
            **kwargs: Dict[str, Any],
        ) -> None:
            super().__init__(*args, **kwargs)
            self.setupUi(self)

    return UiFileWidget



import enum

__copyright__ = "Copyright 2020-2021, Gispo Ltd"
__license__ = "GPL version 3"
__email__ = "info@gispo.fi"
__revision__ = "$Format:%H$"

import logging
from typing import List, Optional, Set, Union

from qgis.core import (
    QgsExpression,
    QgsExpressionContext,
    QgsExpressionContextScope,
    QgsExpressionContextUtils,
    QgsFeature,
    QgsGeometry,
    QgsMapLayer,
    QgsVectorLayer,
    QgsWkbTypes,
)

from .custom_logging import bar_msg
from .exceptions import QgsPluginExpressionException
from .i18n import tr

try:
    from qgis.core import QgsUnitTypes, QgsVectorLayerTemporalProperties
except ImportError:
    QgsVectorLayerTemporalProperties = QgsUnitTypes = None

LOGGER = logging.getLogger(__name__)

POINT_TYPES = {
    QgsWkbTypes.Point,
    QgsWkbTypes.MultiPoint,
}

LINE_TYPES = {
    QgsWkbTypes.LineString,
    QgsWkbTypes.MultiLineString,
}
POLYGON_TYPES = {
    QgsWkbTypes.Polygon,
    QgsWkbTypes.MultiPolygon,
    QgsWkbTypes.CurvePolygon,
}


@enum.unique
class LayerType(enum.Enum):
    Point = {"wkb_types": POINT_TYPES}
    Line = {"wkb_types": LINE_TYPES}
    Polygon = {"wkb_types": POLYGON_TYPES}
    Unknown = {"wkb_types": set()}  # type: ignore

    @staticmethod
    def from_wkb_type(wkb_type: int) -> "LayerType":
        for l_type in LayerType:
            if QgsWkbTypes.flatType(wkb_type) in l_type.wkb_types:
                return l_type
        return LayerType.Unknown

    @staticmethod
    def from_layer(layer: QgsVectorLayer) -> "LayerType":
        return LayerType.from_wkb_type(layer.wkbType())

    @staticmethod
    def from_geometry(geometry: QgsGeometry) -> "LayerType":
        return LayerType.from_wkb_type(geometry.wkbType())

    @property
    def wkb_types(self) -> Set[QgsWkbTypes.GeometryType]:
        return self.value["wkb_types"]


def set_temporal_settings(
    layer: QgsVectorLayer,
    dt_field: str,
    time_step: int,
    unit: "QgsUnitTypes.TemporalUnit" = None,
) -> None:
    """
    Set temporal settings for vector layer temporal range for raster layer
    :param layer: raster layer
    :param dt_field: name of the date time field
    :param time_step: time step in some QgsUnitTypes.TemporalUnit
    :param unit: QgsUnitTypes.TemporalUnit
    """
    if unit is None:
        unit = QgsUnitTypes.TemporalMinutes
    mode = QgsVectorLayerTemporalProperties.ModeFeatureDateTimeInstantFromField
    tprops: QgsVectorLayerTemporalProperties = layer.temporalProperties()
    tprops.setMode(mode)
    tprops.setStartField(dt_field)
    tprops.setFixedDuration(time_step)
    tprops.setDurationUnits(unit)
    tprops.setIsActive(True)


def evaluate_expressions(
    exp: QgsExpression,
    feature: Optional[QgsFeature] = None,
    layer: Optional[QgsMapLayer] = None,
    context_scopes: Optional[List[QgsExpressionContextScope]] = None,
) -> Union[bool, int, str, float, None]:
    """
    Evaluate a QGIS expression
    :param exp: QGIS expression
    :param feature: Optional QgsFeature
    :param layer: Optional QgsMapLayer
    :param context_scopes: Optional list of QgsExpressionContextScopes
    :return: evaluated value of the expression
    """
    context = QgsExpressionContext()
    scopes = context_scopes if context_scopes is not None else []

    if layer:
        scopes.append(QgsExpressionContextUtils.layerScope(layer))
    context.appendScopes(scopes)
    if feature:
        context.setFeature(feature)

    value = exp.evaluate(context)
    if exp.hasParserError():
        raise QgsPluginExpressionException(bar_msg=bar_msg(exp.parserErrorString()))

    if exp.hasEvalError():
        raise QgsPluginExpressionException(bar_msg=bar_msg(exp.evalErrorString()))
    return value


def get_field_index(layer: QgsVectorLayer, field_name: str) -> int:
    """
    Get field index if exists
    :param layer: QgsVectorLayer
    :param field_name: name of the field
    :return: index of the field
    """
    field_index = layer.fields().indexFromName(field_name)
    if field_index == -1:
        raise KeyError(
            tr(
                "Field name {} does not exist in layer {}",
                field_name,
                layer.name(),
            )
        )
    return field_index



from typing import Any, Optional, Union

from qgis.core import QgsExpressionContextUtils, QgsProject, QgsSettings
from qgis.PyQt.QtCore import QVariant

from .exceptions import QgsPluginInvalidProjectSetting
from .resources import plugin_name


def setting_key(*args: str) -> str:
    """
    Get QGIS setting key

    :param args List of path elements e.g. ['img', 'logos', 'image.png']
    """
    return "/" + "/".join((plugin_name(), *map(str, args)))


def get_setting(
    key: str,
    default: Optional[Any] = None,
    typehint: Optional[type] = None,
    internal: bool = True,
    section: int = QgsSettings.NoSection,
) -> Union[QVariant, str]:
    """
    Get QGIS setting value plugin

    :param key: Key for the setting
    :param default: Optional default value
    :param typehint: Type hint
    :param internal: Whether to search from only plugin settings or all
    :param section: Section argument can be used to get a value from
    a specific Section.
    """
    s = QgsSettings()
    kwargs = {"defaultValue": default, "section": section}

    if typehint is not None:
        kwargs["type"] = typehint
    return s.value(setting_key(key) if internal else key, **kwargs)


def set_setting(
    key: str,
    value: Union[str, int, float, bool],
    internal: bool = True,
    section: int = QgsSettings.NoSection,
) -> bool:
    """
    Set a value in the QgsSetting

    :param key: Key for the setting
    :param value: Value for the setting
    :param internal: Whether to search from only plugin settings or all
    :param section: Section argument can be used to set a value to a specific Section
    """
    qs = QgsSettings()
    return qs.setValue(setting_key(key) if internal else key, value, section)


def get_project_setting(
    key: str,
    default: Optional[Any] = None,
    typehint: Optional[type] = None,
    internal: bool = True,
) -> Union[QVariant, str, None]:
    """
    Get QGIS project setting value

    :param key: Key for the setting
    :param default: Optional default value
    :param typehint: Type hint
    :param internal: Whether to search from only plugin settings or all
    :return: Value if conversion is successful, else None
    """
    proj = QgsProject.instance()

    if not internal:
        value = QgsExpressionContextUtils.projectScope(proj).variable(key)
        return value if value is not None else default

    args = [plugin_name(), key]
    if default is not None:
        args.append(default)

    value = None
    conversion_ok = False
    if typehint is not None and typehint is not str:
        try:
            if typehint is int:
                value, conversion_ok = proj.readNumEntry(*args)
            elif typehint is bool:
                value, conversion_ok = proj.readBoolEntry(*args)
            elif typehint is list:
                value, conversion_ok = proj.readListEntry(*args)
        except TypeError as e:
            raise QgsPluginInvalidProjectSetting(str(e))
    else:
        value, conversion_ok = proj.readEntry(*args)
    return value if conversion_ok else default


def set_project_setting(
    key: str, value: Union[str, int, float, bool], internal: bool = True
) -> bool:
    """
    Set a value in the QGIS project settings

    :param key: Key for the setting
    :param value: Value for the setting
    :param internal: Whether to search from only plugin settings or all
    """
    proj = QgsProject.instance()
    if internal:
        return proj.writeEntry(plugin_name(), key, value)
    else:
        QgsExpressionContextUtils.setProjectVariable(proj, key, value)
        return True


def parse_value(value: Union[QVariant, str]) -> Union[None, str, bool]:
    """
    Parse QSettings value

    :param value: QVariant
    """
    str_value = str(value)
    val: Union[None, str, bool] = str_value
    if str_value == "NULL":
        val = None
    elif str_value == "true":
        val = True
    elif str_value == "false":
        val = False
    return val



from qgis.core import QgsRectangle

from ..tools.misc_utils import extent_to_bbox


def test_extent_to_bbox():
    extent = QgsRectangle(1, 2, 3, 4)
    bbox = extent_to_bbox(extent, precision=1)
    assert bbox == "1.0,2.0,3.0,4.0"



# flake8: noqa ANN201, ANN001, ANN204
"""QCombobox with checkbox for selecting multiple items."""
from qgis.core import QgsMapLayer
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtGui import QStandardItem, QStandardItemModel
from qgis.PyQt.QtWidgets import QStyledItemDelegate

__copyright__ = "Copyright 2019, 3Liz"
__license__ = "GPL version 3"
__email__ = "info@3liz.org"
__revision__ = "$Format:%H$"


class CheckableComboBox:

    """Basic QCombobox with selectable items."""

    def __init__(self, combobox, select_all=None):
        """Constructor."""
        self.combo = combobox
        self.combo.setEditable(True)
        self.combo.lineEdit().setReadOnly(True)
        self.model = QStandardItemModel(self.combo)
        self.combo.setModel(self.model)
        self.combo.setItemDelegate(QStyledItemDelegate())
        self.model.itemChanged.connect(self.combo_changed)
        self.combo.lineEdit().textChanged.connect(self.text_changed)
        if select_all:
            self.select_all = select_all
            self.select_all.clicked.connect(self.select_all_clicked)

    def select_all_clicked(self):
        for item in self.model.findItems("*", Qt.MatchWildcard):
            item.setCheckState(Qt.Checked)

    def append_row(self, item: QStandardItem):
        """Add an item to the combobox."""
        item.setEnabled(True)
        item.setCheckable(True)
        item.setSelectable(False)
        self.model.appendRow(item)

    def combo_changed(self):
        """Slot when the combo has changed."""
        self.text_changed(None)

    def selected_items(self) -> list:
        checked_items = []
        for item in self.model.findItems("*", Qt.MatchWildcard):
            if item.checkState() == Qt.Checked:
                checked_items.append(item.data())
        return checked_items

    def set_selected_items(self, items):
        for item in self.model.findItems("*", Qt.MatchWildcard):
            checked = item.data() in items
            item.setCheckState(Qt.Checked if checked else Qt.Unchecked)

    def text_changed(self, text):
        """Update the preview with all selected items, separated by a comma."""
        label = ", ".join(self.selected_items())
        if text != label:
            self.combo.setEditText(label)


class CheckableFieldComboBox(CheckableComboBox):
    def __init__(self, combobox, select_all=None):
        self.layer = None
        super().__init__(combobox, select_all)

    def set_layer(self, layer):
        self.model.clear()

        if not layer:
            return
        if layer.type() != QgsMapLayer.VectorLayer:
            return

        self.layer = layer

        for i, field in enumerate(self.layer.fields()):
            alias = field.alias()
            if alias:
                name = "{} ({})".format(field.name(), alias)
            else:
                name = field.name()
            item = QStandardItem(name)
            item.setData(field.name())
            item.setIcon(self.layer.fields().iconForField(i))
            self.append_row(item)
